import numpy as np
import copy
import os
import time
import sys
from abc import abstractmethod
import abc
if sys.version_info >= (3, 4):
  ABC = abc.ABC
else:
  ABC = abc.ABCMeta('ABC', (), {})

from enum import Enum

from pybullet_envs.deep_mimic.learning.path import *
from pybullet_envs.deep_mimic.learning.exp_params import ExpParams
from pybullet_envs.deep_mimic.learning.normalizer import Normalizer
from pybullet_envs.deep_mimic.learning.replay_buffer import ReplayBuffer
from pybullet_utils.logger import Logger
import pybullet_utils.mpi_util as MPIUtil
import pybullet_utils.math_util as MathUtil


class RLAgent(ABC):

  class Mode(Enum):
    TRAIN = 0
    TEST = 1
    TRAIN_END = 2

  NAME = "None"

  UPDATE_PERIOD_KEY = "UpdatePeriod"
  ITERS_PER_UPDATE = "ItersPerUpdate"
  DISCOUNT_KEY = "Discount"
  MINI_BATCH_SIZE_KEY = "MiniBatchSize"
  REPLAY_BUFFER_SIZE_KEY = "ReplayBufferSize"
  INIT_SAMPLES_KEY = "InitSamples"
  NORMALIZER_SAMPLES_KEY = "NormalizerSamples"

  OUTPUT_ITERS_KEY = "OutputIters"
  INT_OUTPUT_ITERS_KEY = "IntOutputIters"
  TEST_EPISODES_KEY = "TestEpisodes"

  EXP_ANNEAL_SAMPLES_KEY = "ExpAnnealSamples"
  EXP_PARAM_BEG_KEY = "ExpParamsBeg"
  EXP_PARAM_END_KEY = "ExpParamsEnd"

  def __init__(self, world, id, json_data):
    self.world = world
    self.id = id
    self.logger = Logger()
    self._mode = self.Mode.TRAIN

    assert self._check_action_space(), \
        Logger.print2("Invalid action space, got {:s}".format(str(self.get_action_space())))

    self._enable_training = True
    self.path = Path()
    self.iter = int(0)
    self.start_time = time.time()
    self._update_counter = 0

    self.update_period = 1.0  # simulated time (seconds) before each training update
    self.iters_per_update = int(1)
    self.discount = 0.95
    self.mini_batch_size = int(32)
    self.replay_buffer_size = int(50000)
    self.init_samples = int(1000)
    self.normalizer_samples = np.inf
    self._local_mini_batch_size = self.mini_batch_size  # batch size for each work for multiprocessing
    self._need_normalizer_update = True
    self._total_sample_count = 0

    self._output_dir = ""
    self._int_output_dir = ""
    self.output_iters = 100
    self.int_output_iters = 100

    self.train_return = 0.0
    self.test_episodes = int(0)
    self.test_episode_count = int(0)
    self.test_return = 0.0
    self.avg_test_return = 0.0

    self.exp_anneal_samples = 320000
    self.exp_params_beg = ExpParams()
    self.exp_params_end = ExpParams()
    self.exp_params_curr = ExpParams()

    self._load_params(json_data)
    self._build_replay_buffer(self.replay_buffer_size)
    self._build_normalizers()
    self._build_bounds()
    self.reset()

    return

  def __str__(self):
    action_space_str = str(self.get_action_space())
    info_str = ""
    info_str += '"ID": {:d},\n "Type": "{:s}",\n "ActionSpace": "{:s}",\n "StateDim": {:d},\n "GoalDim": {:d},\n "ActionDim": {:d}'.format(
        self.id, self.NAME, action_space_str[action_space_str.rfind('.') + 1:],
        self.get_state_size(), self.get_goal_size(), self.get_action_size())
    return "{\n" + info_str + "\n}"

  def get_output_dir(self):
    return self._output_dir

  def set_output_dir(self, out_dir):
    self._output_dir = out_dir
    if (self._output_dir != ""):
      self.logger.configure_output_file(out_dir + "/agent" + str(self.id) + "_log.txt")
    return

  output_dir = property(get_output_dir, set_output_dir)

  def get_int_output_dir(self):
    return self._int_output_dir

  def set_int_output_dir(self, out_dir):
    self._int_output_dir = out_dir
    return

  int_output_dir = property(get_int_output_dir, set_int_output_dir)

  def reset(self):
    self.path.clear()
    return

  def update(self, timestep):
    if self.need_new_action():
      #print("update_new_action!!!")
      self._update_new_action()

    if (self._mode == self.Mode.TRAIN and self.enable_training):
      self._update_counter += timestep

      while self._update_counter >= self.update_period:
        self._train()
        self._update_exp_params()
        self.world.env.set_sample_count(self._total_sample_count)
        self._update_counter -= self.update_period

    return

  def end_episode(self):
    if (self.path.pathlength() > 0):
      self._end_path()

      if (self._mode == self.Mode.TRAIN or self._mode == self.Mode.TRAIN_END):
        if (self.enable_training and self.path.pathlength() > 0):
          self._store_path(self.path)
      elif (self._mode == self.Mode.TEST):
        self._update_test_return(self.path)
      else:
        assert False, Logger.print2("Unsupported RL agent mode" + str(self._mode))

      self._update_mode()
    return

  def has_goal(self):
    return self.get_goal_size() > 0

  def predict_val(self):
    return 0

  def get_enable_training(self):
    return self._enable_training

  def set_enable_training(self, enable):
    print("set_enable_training=", enable)
    self._enable_training = enable
    if (self._enable_training):
      self.reset()
    return

  enable_training = property(get_enable_training, set_enable_training)

  def enable_testing(self):
    return self.test_episodes > 0

  def get_name(self):
    return self.NAME

  @abstractmethod
  def save_model(self, out_path):
    pass

  @abstractmethod
  def load_model(self, in_path):
    pass

  @abstractmethod
  def _decide_action(self, s, g):
    pass

  @abstractmethod
  def _get_output_path(self):
    pass

  @abstractmethod
  def _get_int_output_path(self):
    pass

  @abstractmethod
  def _train_step(self):
    pass

  @abstractmethod
  def _check_action_space(self):
    pass

  def get_action_space(self):
    return self.world.env.get_action_space(self.id)

  def get_state_size(self):
    return self.world.env.get_state_size(self.id)

  def get_goal_size(self):
    return self.world.env.get_goal_size(self.id)

  def get_action_size(self):
    return self.world.env.get_action_size(self.id)

  def get_num_actions(self):
    return self.world.env.get_num_actions(self.id)

  def need_new_action(self):
    return self.world.env.need_new_action(self.id)

  def _build_normalizers(self):
    self.s_norm = Normalizer(self.get_state_size(),
                             self.world.env.build_state_norm_groups(self.id))
    self.s_norm.set_mean_std(-self.world.env.build_state_offset(self.id),
                             1 / self.world.env.build_state_scale(self.id))

    self.g_norm = Normalizer(self.get_goal_size(), self.world.env.build_goal_norm_groups(self.id))
    self.g_norm.set_mean_std(-self.world.env.build_goal_offset(self.id),
                             1 / self.world.env.build_goal_scale(self.id))

    self.a_norm = Normalizer(self.world.env.get_action_size())
    self.a_norm.set_mean_std(-self.world.env.build_action_offset(self.id),
                             1 / self.world.env.build_action_scale(self.id))
    return

  def _build_bounds(self):
    self.a_bound_min = self.world.env.build_action_bound_min(self.id)
    self.a_bound_max = self.world.env.build_action_bound_max(self.id)
    return

  def _load_params(self, json_data):
    if (self.UPDATE_PERIOD_KEY in json_data):
      self.update_period = int(json_data[self.UPDATE_PERIOD_KEY])

    if (self.ITERS_PER_UPDATE in json_data):
      self.iters_per_update = int(json_data[self.ITERS_PER_UPDATE])

    if (self.DISCOUNT_KEY in json_data):
      self.discount = json_data[self.DISCOUNT_KEY]

    if (self.MINI_BATCH_SIZE_KEY in json_data):
      self.mini_batch_size = int(json_data[self.MINI_BATCH_SIZE_KEY])

    if (self.REPLAY_BUFFER_SIZE_KEY in json_data):
      self.replay_buffer_size = int(json_data[self.REPLAY_BUFFER_SIZE_KEY])

    if (self.INIT_SAMPLES_KEY in json_data):
      self.init_samples = int(json_data[self.INIT_SAMPLES_KEY])

    if (self.NORMALIZER_SAMPLES_KEY in json_data):
      self.normalizer_samples = int(json_data[self.NORMALIZER_SAMPLES_KEY])

    if (self.OUTPUT_ITERS_KEY in json_data):
      self.output_iters = json_data[self.OUTPUT_ITERS_KEY]

    if (self.INT_OUTPUT_ITERS_KEY in json_data):
      self.int_output_iters = json_data[self.INT_OUTPUT_ITERS_KEY]

    if (self.TEST_EPISODES_KEY in json_data):
      self.test_episodes = int(json_data[self.TEST_EPISODES_KEY])

    if (self.EXP_ANNEAL_SAMPLES_KEY in json_data):
      self.exp_anneal_samples = json_data[self.EXP_ANNEAL_SAMPLES_KEY]

    if (self.EXP_PARAM_BEG_KEY in json_data):
      self.exp_params_beg.load(json_data[self.EXP_PARAM_BEG_KEY])

    if (self.EXP_PARAM_END_KEY in json_data):
      self.exp_params_end.load(json_data[self.EXP_PARAM_END_KEY])

    num_procs = MPIUtil.get_num_procs()
    self._local_mini_batch_size = int(np.ceil(self.mini_batch_size / num_procs))
    self._local_mini_batch_size = np.maximum(self._local_mini_batch_size, 1)
    self.mini_batch_size = self._local_mini_batch_size * num_procs

    assert (self.exp_params_beg.noise == self.exp_params_end.noise)  # noise std should not change
    self.exp_params_curr = copy.deepcopy(self.exp_params_beg)
    self.exp_params_end.noise = self.exp_params_beg.noise

    self._need_normalizer_update = self.normalizer_samples > 0

    return

  def _record_state(self):
    s = self.world.env.record_state(self.id)
    return s

  def _record_goal(self):
    g = self.world.env.record_goal(self.id)
    return g

  def _record_reward(self):
    r = self.world.env.calc_reward(self.id)
    return r

  def _apply_action(self, a):
    self.world.env.set_action(self.id, a)
    return

  def _record_flags(self):
    return int(0)

  def _is_first_step(self):
    return len(self.path.states) == 0

  def _end_path(self):
    s = self._record_state()
    g = self._record_goal()
    r = self._record_reward()

    self.path.rewards.append(r)
    self.path.states.append(s)
    self.path.goals.append(g)
    self.path.terminate = self.world.env.check_terminate(self.id)

    return

  def _update_new_action(self):
    #print("_update_new_action!")
    s = self._record_state()
    #np.savetxt("pb_record_state_s.csv", s, delimiter=",")
    g = self._record_goal()

    if not (self._is_first_step()):
      r = self._record_reward()
      self.path.rewards.append(r)

    a, logp = self._decide_action(s=s, g=g)
    assert len(np.shape(a)) == 1
    assert len(np.shape(logp)) <= 1

    flags = self._record_flags()
    self._apply_action(a)

    self.path.states.append(s)
    self.path.goals.append(g)
    self.path.actions.append(a)
    self.path.logps.append(logp)
    self.path.flags.append(flags)

    if self._enable_draw():
      self._log_val(s, g)

    return

  def _update_exp_params(self):
    lerp = float(self._total_sample_count) / self.exp_anneal_samples
    lerp = np.clip(lerp, 0.0, 1.0)
    self.exp_params_curr = self.exp_params_beg.lerp(self.exp_params_end, lerp)
    return

  def _update_test_return(self, path):
    path_reward = path.calc_return()
    self.test_return += path_reward
    self.test_episode_count += 1
    return

  def _update_mode(self):
    if (self._mode == self.Mode.TRAIN):
      self._update_mode_train()
    elif (self._mode == self.Mode.TRAIN_END):
      self._update_mode_train_end()
    elif (self._mode == self.Mode.TEST):
      self._update_mode_test()
    else:
      assert False, Logger.print2("Unsupported RL agent mode" + str(self._mode))
    return

  def _update_mode_train(self):
    return

  def _update_mode_train_end(self):
    self._init_mode_test()
    return

  def _update_mode_test(self):
    if (self.test_episode_count * MPIUtil.get_num_procs() >= self.test_episodes):
      global_return = MPIUtil.reduce_sum(self.test_return)
      global_count = MPIUtil.reduce_sum(self.test_episode_count)
      avg_return = global_return / global_count
      self.avg_test_return = avg_return

      if self.enable_training:
        self._init_mode_train()
    return

  def _init_mode_train(self):
    self._mode = self.Mode.TRAIN
    self.world.env.set_mode(self._mode)
    return

  def _init_mode_train_end(self):
    self._mode = self.Mode.TRAIN_END
    return

  def _init_mode_test(self):
    self._mode = self.Mode.TEST
    self.test_return = 0.0
    self.test_episode_count = 0
    self.world.env.set_mode(self._mode)
    return

  def _enable_output(self):
    return MPIUtil.is_root_proc() and self.output_dir != ""

  def _enable_int_output(self):
    return MPIUtil.is_root_proc() and self.int_output_dir != ""

  def _calc_val_bounds(self, discount):
    r_min = self.world.env.get_reward_min(self.id)
    r_max = self.world.env.get_reward_max(self.id)
    assert (r_min <= r_max)

    val_min = r_min / (1.0 - discount)
    val_max = r_max / (1.0 - discount)
    return val_min, val_max

  def _calc_val_offset_scale(self, discount):
    val_min, val_max = self._calc_val_bounds(discount)
    val_offset = 0
    val_scale = 1

    if (np.isfinite(val_min) and np.isfinite(val_max)):
      val_offset = -0.5 * (val_max + val_min)
      val_scale = 2 / (val_max - val_min)

    return val_offset, val_scale

  def _calc_term_vals(self, discount):
    r_fail = self.world.env.get_reward_fail(self.id)
    r_succ = self.world.env.get_reward_succ(self.id)

    r_min = self.world.env.get_reward_min(self.id)
    r_max = self.world.env.get_reward_max(self.id)
    assert (r_fail <= r_max and r_fail >= r_min)
    assert (r_succ <= r_max and r_succ >= r_min)
    assert (not np.isinf(r_fail))
    assert (not np.isinf(r_succ))

    if (discount == 0):
      val_fail = 0
      val_succ = 0
    else:
      val_fail = r_fail / (1.0 - discount)
      val_succ = r_succ / (1.0 - discount)

    return val_fail, val_succ

  def _update_iter(self, iter):
    if (self._enable_output() and self.iter % self.output_iters == 0):
      output_path = self._get_output_path()
      output_dir = os.path.dirname(output_path)
      if not os.path.exists(output_dir):
        os.makedirs(output_dir)
      self.save_model(output_path)

    if (self._enable_int_output() and self.iter % self.int_output_iters == 0):
      int_output_path = self._get_int_output_path()
      int_output_dir = os.path.dirname(int_output_path)
      if not os.path.exists(int_output_dir):
        os.makedirs(int_output_dir)
      self.save_model(int_output_path)

    self.iter = iter
    return

  def _enable_draw(self):
    return self.world.env.enable_draw

  def _log_val(self, s, g):
    pass

  def _build_replay_buffer(self, buffer_size):
    num_procs = MPIUtil.get_num_procs()
    buffer_size = int(buffer_size / num_procs)
    self.replay_buffer = ReplayBuffer(buffer_size=buffer_size)
    self.replay_buffer_initialized = False
    return

  def _store_path(self, path):
    path_id = self.replay_buffer.store(path)
    valid_path = path_id != MathUtil.INVALID_IDX

    if valid_path:
      self.train_return = path.calc_return()

      if self._need_normalizer_update:
        self._record_normalizers(path)

    return path_id

  def _record_normalizers(self, path):
    states = np.array(path.states)
    self.s_norm.record(states)

    if self.has_goal():
      goals = np.array(path.goals)
      self.g_norm.record(goals)

    return

  def _update_normalizers(self):
    self.s_norm.update()

    if self.has_goal():
      self.g_norm.update()
    return

  def _train(self):
    samples = self.replay_buffer.total_count
    self._total_sample_count = int(MPIUtil.reduce_sum(samples))
    end_training = False

    if (self.replay_buffer_initialized):
      if (self._valid_train_step()):
        prev_iter = self.iter
        iters = self._get_iters_per_update()
        avg_train_return = MPIUtil.reduce_avg(self.train_return)

        for i in range(iters):
          curr_iter = self.iter
          wall_time = time.time() - self.start_time
          wall_time /= 60 * 60  # store time in hours

          has_goal = self.has_goal()
          s_mean = np.mean(self.s_norm.mean)
          s_std = np.mean(self.s_norm.std)
          g_mean = np.mean(self.g_norm.mean) if has_goal else 0
          g_std = np.mean(self.g_norm.std) if has_goal else 0

          self.logger.log_tabular("Iteration", self.iter)
          self.logger.log_tabular("Wall_Time", wall_time)
          self.logger.log_tabular("Samples", self._total_sample_count)
          self.logger.log_tabular("Train_Return", avg_train_return)
          self.logger.log_tabular("Test_Return", self.avg_test_return)
          self.logger.log_tabular("State_Mean", s_mean)
          self.logger.log_tabular("State_Std", s_std)
          self.logger.log_tabular("Goal_Mean", g_mean)
          self.logger.log_tabular("Goal_Std", g_std)
          self._log_exp_params()

          self._update_iter(self.iter + 1)
          self._train_step()

          Logger.print2("Agent " + str(self.id))
          self.logger.print_tabular()
          Logger.print2("")

          if (self._enable_output() and curr_iter % self.int_output_iters == 0):
            self.logger.dump_tabular()

        if (prev_iter // self.int_output_iters != self.iter // self.int_output_iters):
          end_training = self.enable_testing()

    else:

      Logger.print2("Agent " + str(self.id))
      Logger.print2("Samples: " + str(self._total_sample_count))
      Logger.print2("")

      if (self._total_sample_count >= self.init_samples):
        self.replay_buffer_initialized = True
        end_training = self.enable_testing()

    if self._need_normalizer_update:
      self._update_normalizers()
      self._need_normalizer_update = self.normalizer_samples > self._total_sample_count

    if end_training:
      self._init_mode_train_end()

    return

  def _get_iters_per_update(self):
    return MPIUtil.get_num_procs() * self.iters_per_update

  def _valid_train_step(self):
    return True

  def _log_exp_params(self):
    self.logger.log_tabular("Exp_Rate", self.exp_params_curr.rate)
    self.logger.log_tabular("Exp_Noise", self.exp_params_curr.noise)
    self.logger.log_tabular("Exp_Temp", self.exp_params_curr.temp)
    return



import numpy as np
try:
  import tensorflow.compat.v1 as tf
except Exception:
  import tensorflow as tf

import copy

from pybullet_envs.deep_mimic.learning.tf_agent import TFAgent
from pybullet_envs.deep_mimic.learning.solvers.mpi_solver import MPISolver
import pybullet_envs.deep_mimic.learning.tf_util as TFUtil
import pybullet_envs.deep_mimic.learning.nets.net_builder as NetBuilder
from pybullet_envs.deep_mimic.learning.tf_normalizer import TFNormalizer
import pybullet_envs.deep_mimic.learning.rl_util as RLUtil
from pybullet_utils.logger import Logger
import pybullet_utils.mpi_util as MPIUtil
import pybullet_utils.math_util as MathUtil
from pybullet_envs.deep_mimic.env.action_space import ActionSpace
from pybullet_envs.deep_mimic.env.env import Env
'''
Policy Gradient Agent
'''


class PGAgent(TFAgent):
  NAME = 'PG'

  ACTOR_NET_KEY = 'ActorNet'
  ACTOR_STEPSIZE_KEY = 'ActorStepsize'
  ACTOR_MOMENTUM_KEY = 'ActorMomentum'
  ACTOR_WEIGHT_DECAY_KEY = 'ActorWeightDecay'
  ACTOR_INIT_OUTPUT_SCALE_KEY = 'ActorInitOutputScale'

  CRITIC_NET_KEY = 'CriticNet'
  CRITIC_STEPSIZE_KEY = 'CriticStepsize'
  CRITIC_MOMENTUM_KEY = 'CriticMomentum'
  CRITIC_WEIGHT_DECAY_KEY = 'CriticWeightDecay'

  EXP_ACTION_FLAG = 1 << 0

  def __init__(self, world, id, json_data):
    self._exp_action = False
    super().__init__(world, id, json_data)
    return

  def reset(self):
    super().reset()
    self._exp_action = False
    return

  def _check_action_space(self):
    action_space = self.get_action_space()
    return action_space == ActionSpace.Continuous

  def _load_params(self, json_data):
    super()._load_params(json_data)
    self.val_min, self.val_max = self._calc_val_bounds(self.discount)
    self.val_fail, self.val_succ = self._calc_term_vals(self.discount)
    return

  def _build_nets(self, json_data):
    assert self.ACTOR_NET_KEY in json_data
    assert self.CRITIC_NET_KEY in json_data

    actor_net_name = json_data[self.ACTOR_NET_KEY]
    critic_net_name = json_data[self.CRITIC_NET_KEY]
    actor_init_output_scale = 1 if (self.ACTOR_INIT_OUTPUT_SCALE_KEY not in json_data
                                   ) else json_data[self.ACTOR_INIT_OUTPUT_SCALE_KEY]

    s_size = self.get_state_size()
    g_size = self.get_goal_size()
    a_size = self.get_action_size()

    # setup input tensors
    self.s_tf = tf.placeholder(tf.float32, shape=[None, s_size], name="s")  # observations
    self.tar_val_tf = tf.placeholder(tf.float32, shape=[None], name="tar_val")  # target value s
    self.adv_tf = tf.placeholder(tf.float32, shape=[None], name="adv")  # advantage
    self.a_tf = tf.placeholder(tf.float32, shape=[None, a_size], name="a")  # target actions
    self.g_tf = tf.placeholder(tf.float32,
                               shape=([None, g_size] if self.has_goal() else None),
                               name="g")  # goals

    with tf.variable_scope('main'):
      with tf.variable_scope('actor'):
        self.actor_tf = self._build_net_actor(actor_net_name, actor_init_output_scale)
      with tf.variable_scope('critic'):
        self.critic_tf = self._build_net_critic(critic_net_name)

    if (self.actor_tf != None):
      Logger.print2('Built actor net: ' + actor_net_name)

    if (self.critic_tf != None):
      Logger.print2('Built critic net: ' + critic_net_name)

    return

  def _build_normalizers(self):
    super()._build_normalizers()
    with self.sess.as_default(), self.graph.as_default(), tf.variable_scope(self.tf_scope):
      with tf.variable_scope(self.RESOURCE_SCOPE):
        val_offset, val_scale = self._calc_val_offset_scale(self.discount)
        self.val_norm = TFNormalizer(self.sess, 'val_norm', 1)
        self.val_norm.set_mean_std(-val_offset, 1.0 / val_scale)
    return

  def _init_normalizers(self):
    super()._init_normalizers()
    with self.sess.as_default(), self.graph.as_default():
      self.val_norm.update()
    return

  def _load_normalizers(self):
    super()._load_normalizers()
    self.val_norm.load()
    return

  def _build_losses(self, json_data):
    actor_weight_decay = 0 if (
        self.ACTOR_WEIGHT_DECAY_KEY not in json_data) else json_data[self.ACTOR_WEIGHT_DECAY_KEY]
    critic_weight_decay = 0 if (
        self.CRITIC_WEIGHT_DECAY_KEY not in json_data) else json_data[self.CRITIC_WEIGHT_DECAY_KEY]

    norm_val_diff = self.val_norm.normalize_tf(self.tar_val_tf) - self.val_norm.normalize_tf(
        self.critic_tf)
    self.critic_loss_tf = 0.5 * tf.reduce_mean(tf.square(norm_val_diff))

    if (critic_weight_decay != 0):
      self.critic_loss_tf += critic_weight_decay * self._weight_decay_loss('main/critic')

    norm_a_mean_tf = self.a_norm.normalize_tf(self.actor_tf)
    norm_a_diff = self.a_norm.normalize_tf(self.a_tf) - norm_a_mean_tf

    self.actor_loss_tf = tf.reduce_sum(tf.square(norm_a_diff), axis=-1)
    self.actor_loss_tf *= self.adv_tf
    self.actor_loss_tf = 0.5 * tf.reduce_mean(self.actor_loss_tf)

    norm_a_bound_min = self.a_norm.normalize(self.a_bound_min)
    norm_a_bound_max = self.a_norm.normalize(self.a_bound_max)
    a_bound_loss = TFUtil.calc_bound_loss(norm_a_mean_tf, norm_a_bound_min, norm_a_bound_max)
    a_bound_loss /= self.exp_params_curr.noise
    self.actor_loss_tf += a_bound_loss

    if (actor_weight_decay != 0):
      self.actor_loss_tf += actor_weight_decay * self._weight_decay_loss('main/actor')

    return

  def _build_solvers(self, json_data):
    actor_stepsize = 0.001 if (
        self.ACTOR_STEPSIZE_KEY not in json_data) else json_data[self.ACTOR_STEPSIZE_KEY]
    actor_momentum = 0.9 if (
        self.ACTOR_MOMENTUM_KEY not in json_data) else json_data[self.ACTOR_MOMENTUM_KEY]
    critic_stepsize = 0.01 if (
        self.CRITIC_STEPSIZE_KEY not in json_data) else json_data[self.CRITIC_STEPSIZE_KEY]
    critic_momentum = 0.9 if (
        self.CRITIC_MOMENTUM_KEY not in json_data) else json_data[self.CRITIC_MOMENTUM_KEY]

    critic_vars = self._tf_vars('main/critic')
    critic_opt = tf.train.MomentumOptimizer(learning_rate=critic_stepsize,
                                            momentum=critic_momentum)
    self.critic_grad_tf = tf.gradients(self.critic_loss_tf, critic_vars)
    self.critic_solver = MPISolver(self.sess, critic_opt, critic_vars)

    actor_vars = self._tf_vars('main/actor')
    actor_opt = tf.train.MomentumOptimizer(learning_rate=actor_stepsize, momentum=actor_momentum)
    self.actor_grad_tf = tf.gradients(self.actor_loss_tf, actor_vars)
    self.actor_solver = MPISolver(self.sess, actor_opt, actor_vars)

    return

  def _build_net_actor(self, net_name, init_output_scale):
    norm_s_tf = self.s_norm.normalize_tf(self.s_tf)
    input_tfs = [norm_s_tf]
    if (self.has_goal()):
      norm_g_tf = self.g_norm.normalize_tf(self.g_tf)
      input_tfs += [norm_g_tf]

    h = NetBuilder.build_net(net_name, input_tfs)
    norm_a_tf = tf.layers.dense(inputs=h,
                                units=self.get_action_size(),
                                activation=None,
                                kernel_initializer=tf.random_uniform_initializer(
                                    minval=-init_output_scale, maxval=init_output_scale))

    a_tf = self.a_norm.unnormalize_tf(norm_a_tf)
    return a_tf

  def _build_net_critic(self, net_name):
    norm_s_tf = self.s_norm.normalize_tf(self.s_tf)
    input_tfs = [norm_s_tf]
    if (self.has_goal()):
      norm_g_tf = self.g_norm.normalize_tf(self.g_tf)
      input_tfs += [norm_g_tf]

    h = NetBuilder.build_net(net_name, input_tfs)
    norm_val_tf = tf.layers.dense(inputs=h,
                                  units=1,
                                  activation=None,
                                  kernel_initializer=TFUtil.xavier_initializer)

    norm_val_tf = tf.reshape(norm_val_tf, [-1])
    val_tf = self.val_norm.unnormalize_tf(norm_val_tf)
    return val_tf

  def _initialize_vars(self):
    super()._initialize_vars()
    self._sync_solvers()
    return

  def _sync_solvers(self):
    self.actor_solver.sync()
    self.critic_solver.sync()
    return

  def _decide_action(self, s, g):
    with self.sess.as_default(), self.graph.as_default():
      self._exp_action = False
      a = self._eval_actor(s, g)[0]
      logp = 0

      if self._enable_stoch_policy():
        # epsilon-greedy
        rand_action = MathUtil.flip_coin(self.exp_params_curr.rate)
        if rand_action:
          norm_exp_noise = np.random.randn(*a.shape)
          norm_exp_noise *= self.exp_params_curr.noise
          exp_noise = norm_exp_noise * self.a_norm.std
          a += exp_noise

          logp = self._calc_action_logp(norm_exp_noise)
          self._exp_action = True

    return a, logp

  def _enable_stoch_policy(self):
    return self.enable_training and (self._mode == self.Mode.TRAIN or
                                     self._mode == self.Mode.TRAIN_END)

  def _eval_actor(self, s, g):
    s = np.reshape(s, [-1, self.get_state_size()])
    g = np.reshape(g, [-1, self.get_goal_size()]) if self.has_goal() else None

    feed = {self.s_tf: s, self.g_tf: g}

    a = self.actor_tf.eval(feed)
    return a

  def _eval_critic(self, s, g):
    with self.sess.as_default(), self.graph.as_default():
      s = np.reshape(s, [-1, self.get_state_size()])
      g = np.reshape(g, [-1, self.get_goal_size()]) if self.has_goal() else None

      feed = {self.s_tf: s, self.g_tf: g}

      val = self.critic_tf.eval(feed)
    return val

  def _record_flags(self):
    flags = int(0)
    if (self._exp_action):
      flags = flags | self.EXP_ACTION_FLAG
    return flags

  def _train_step(self):
    super()._train_step()

    critic_loss = self._update_critic()
    actor_loss = self._update_actor()
    critic_loss = MPIUtil.reduce_avg(critic_loss)
    actor_loss = MPIUtil.reduce_avg(actor_loss)

    critic_stepsize = self.critic_solver.get_stepsize()
    actor_stepsize = self.actor_solver.get_stepsize()

    self.logger.log_tabular('Critic_Loss', critic_loss)
    self.logger.log_tabular('Critic_Stepsize', critic_stepsize)
    self.logger.log_tabular('Actor_Loss', actor_loss)
    self.logger.log_tabular('Actor_Stepsize', actor_stepsize)

    return

  def _update_critic(self):
    idx = self.replay_buffer.sample(self._local_mini_batch_size)
    s = self.replay_buffer.get('states', idx)
    g = self.replay_buffer.get('goals', idx) if self.has_goal() else None

    tar_V = self._calc_updated_vals(idx)
    tar_V = np.clip(tar_V, self.val_min, self.val_max)

    feed = {self.s_tf: s, self.g_tf: g, self.tar_val_tf: tar_V}

    loss, grads = self.sess.run([self.critic_loss_tf, self.critic_grad_tf], feed)
    self.critic_solver.update(grads)
    return loss

  def _update_actor(self):
    key = self.EXP_ACTION_FLAG
    idx = self.replay_buffer.sample_filtered(self._local_mini_batch_size, key)
    has_goal = self.has_goal()

    s = self.replay_buffer.get('states', idx)
    g = self.replay_buffer.get('goals', idx) if has_goal else None
    a = self.replay_buffer.get('actions', idx)

    V_new = self._calc_updated_vals(idx)
    V_old = self._eval_critic(s, g)
    adv = V_new - V_old

    feed = {self.s_tf: s, self.g_tf: g, self.a_tf: a, self.adv_tf: adv}

    loss, grads = self.sess.run([self.actor_loss_tf, self.actor_grad_tf], feed)
    self.actor_solver.update(grads)

    return loss

  def _calc_updated_vals(self, idx):
    r = self.replay_buffer.get('rewards', idx)

    if self.discount == 0:
      new_V = r
    else:
      next_idx = self.replay_buffer.get_next_idx(idx)
      s_next = self.replay_buffer.get('states', next_idx)
      g_next = self.replay_buffer.get('goals', next_idx) if self.has_goal() else None

      is_end = self.replay_buffer.is_path_end(idx)
      is_fail = self.replay_buffer.check_terminal_flag(idx, Env.Terminate.Fail)
      is_succ = self.replay_buffer.check_terminal_flag(idx, Env.Terminate.Succ)
      is_fail = np.logical_and(is_end, is_fail)
      is_succ = np.logical_and(is_end, is_succ)

      V_next = self._eval_critic(s_next, g_next)
      V_next[is_fail] = self.val_fail
      V_next[is_succ] = self.val_succ

      new_V = r + self.discount * V_next
    return new_V

  def _calc_action_logp(self, norm_action_deltas):
    # norm action delta are for the normalized actions (scaled by self.a_norm.std)
    stdev = self.exp_params_curr.noise
    assert stdev > 0

    a_size = self.get_action_size()
    logp = -0.5 / (stdev * stdev) * np.sum(np.square(norm_action_deltas), axis=-1)
    logp += -0.5 * a_size * np.log(2 * np.pi)
    logp += -a_size * np.log(stdev)
    return logp

  def _log_val(self, s, g):
    val = self._eval_critic(s, g)
    norm_val = self.val_norm.normalize(val)
    self.world.env.log_val(self.id, norm_val[0])
    return

  def _build_replay_buffer(self, buffer_size):
    super()._build_replay_buffer(buffer_size)
    self.replay_buffer.add_filter_key(self.EXP_ACTION_FLAG)
    return



import pybullet_data
from pybullet_utils.arg_parser import ArgParser
from pybullet_utils.logger import Logger
import sys


def build_arg_parser(args):
  arg_parser = ArgParser()
  arg_parser.load_args(args)

  arg_file = arg_parser.parse_string('arg_file', '')
  if (arg_file != ''):
    path = pybullet_data.getDataPath() + "/args/" + arg_file
    succ = arg_parser.load_file(path)
    Logger.print2(arg_file)
    assert succ, Logger.print2('Failed to	load args	from:	' + arg_file)

  return arg_parser


args = sys.argv[1:]
arg_parser = build_arg_parser(args)
motion_file = arg_parser.parse_string("motion_file")
print("motion_file=", motion_file)
bodies = arg_parser.parse_ints("fall_contact_bodies")
print("bodies=", bodies)
int_output_path = arg_parser.parse_string("int_output_path")
print("int_output_path=", int_output_path)



import time
import numpy as np
import math

useNullSpace = 1
ikSolver = 0
pandaEndEffectorIndex = 11 #8
pandaNumDofs = 7

ll = [-7]*pandaNumDofs
#upper limits for null space (todo: set them to proper range)
ul = [7]*pandaNumDofs
#joint ranges for null space (todo: set them to proper range)
jr = [7]*pandaNumDofs
#restposes for null space
jointPositions=[0.98, 0.458, 0.31, -2.24, -0.30, 2.66, 2.32, 0.02, 0.02]
rp = jointPositions

class PandaSim(object):
  def __init__(self, bullet_client, offset):
    self.bullet_client = bullet_client
    self.bullet_client.setPhysicsEngineParameter(solverResidualThreshold=0)
    self.offset = np.array(offset)
    
    #print("offset=",offset)
    flags = self.bullet_client.URDF_ENABLE_CACHED_GRAPHICS_SHAPES
    self.legos=[]
    
    self.bullet_client.loadURDF("tray/traybox.urdf", [0+offset[0], 0+offset[1], -0.6+offset[2]], [-0.5, -0.5, -0.5, 0.5], flags=flags)
    self.legos.append(self.bullet_client.loadURDF("lego/lego.urdf",np.array([0.1, 0.3, -0.5])+self.offset, flags=flags))
    self.bullet_client.changeVisualShape(self.legos[0],-1,rgbaColor=[1,0,0,1])
    self.legos.append(self.bullet_client.loadURDF("lego/lego.urdf",np.array([-0.1, 0.3, -0.5])+self.offset, flags=flags))
    self.legos.append(self.bullet_client.loadURDF("lego/lego.urdf",np.array([0.1, 0.3, -0.7])+self.offset, flags=flags))
    self.sphereId = self.bullet_client.loadURDF("sphere_small.urdf",np.array( [0, 0.3, -0.6])+self.offset, flags=flags)
    self.bullet_client.loadURDF("sphere_small.urdf",np.array( [0, 0.3, -0.5])+self.offset, flags=flags)
    self.bullet_client.loadURDF("sphere_small.urdf",np.array( [0, 0.3, -0.7])+self.offset, flags=flags)
    orn=[-0.707107, 0.0, 0.0, 0.707107]#p.getQuaternionFromEuler([-math.pi/2,math.pi/2,0])
    eul = self.bullet_client.getEulerFromQuaternion([-0.5, -0.5, -0.5, 0.5])
    self.panda = self.bullet_client.loadURDF("franka_panda/panda.urdf", np.array([0,0,0])+self.offset, orn, useFixedBase=True, flags=flags)
    index = 0
    self.state = 0
    self.control_dt = 1./240.
    self.finger_target = 0
    self.gripper_height = 0.2
    #create a constraint to keep the fingers centered
    c = self.bullet_client.createConstraint(self.panda,
                       9,
                       self.panda,
                       10,
                       jointType=self.bullet_client.JOINT_GEAR,
                       jointAxis=[1, 0, 0],
                       parentFramePosition=[0, 0, 0],
                       childFramePosition=[0, 0, 0])
    self.bullet_client.changeConstraint(c, gearRatio=-1, erp=0.1, maxForce=50)
 
    for j in range(self.bullet_client.getNumJoints(self.panda)):
      self.bullet_client.changeDynamics(self.panda, j, linearDamping=0, angularDamping=0)
      info = self.bullet_client.getJointInfo(self.panda, j)
      #print("info=",info)
      jointName = info[1]
      jointType = info[2]
      if (jointType == self.bullet_client.JOINT_PRISMATIC):
        
        self.bullet_client.resetJointState(self.panda, j, jointPositions[index]) 
        index=index+1
      if (jointType == self.bullet_client.JOINT_REVOLUTE):
        self.bullet_client.resetJointState(self.panda, j, jointPositions[index]) 
        index=index+1
    self.t = 0.
  def reset(self):
    pass

  def update_state(self):
    keys = self.bullet_client.getKeyboardEvents()
    if len(keys)>0:
      for k,v in keys.items():
        if v&self.bullet_client.KEY_WAS_TRIGGERED:
          if (k==ord('1')):
            self.state = 1
          if (k==ord('2')):
            self.state = 2
          if (k==ord('3')):
            self.state = 3
          if (k==ord('4')):
            self.state = 4
          if (k==ord('5')):
                self.state = 5
          if (k==ord('6')):
                self.state = 6
        if v&self.bullet_client.KEY_WAS_RELEASED:
            self.state = 0
  def step(self):
    if self.state==6:
      self.finger_target = 0.01
    if self.state==5:
      self.finger_target = 0.04 
    self.bullet_client.submitProfileTiming("step")
    self.update_state()
    #print("self.state=",self.state)
    #print("self.finger_target=",self.finger_target)
    alpha = 0.9 #0.99
    if self.state==1 or self.state==2 or self.state==3 or self.state==4 or self.state==7:
      #gripper_height = 0.034
      self.gripper_height = alpha * self.gripper_height + (1.-alpha)*0.03
      if self.state == 2 or self.state == 3 or self.state == 7:
        self.gripper_height = alpha * self.gripper_height + (1.-alpha)*0.2
      
      t = self.t
      self.t += self.control_dt
      pos = [self.offset[0]+0.2 * math.sin(1.5 * t), self.offset[1]+self.gripper_height, self.offset[2]+-0.6 + 0.1 * math.cos(1.5 * t)]
      if self.state == 3 or self.state== 4:
        pos, o = self.bullet_client.getBasePositionAndOrientation(self.legos[0])
        pos = [pos[0], self.gripper_height, pos[2]]
        self.prev_pos = pos
      if self.state == 7:
        pos = self.prev_pos
        diffX = pos[0] - self.offset[0]
        diffZ = pos[2] - (self.offset[2]-0.6)
        self.prev_pos = [self.prev_pos[0] - diffX*0.1, self.prev_pos[1], self.prev_pos[2]-diffZ*0.1]

      	
      orn = self.bullet_client.getQuaternionFromEuler([math.pi/2.,0.,0.])
      self.bullet_client.submitProfileTiming("IK")
      jointPoses = self.bullet_client.calculateInverseKinematics(self.panda,pandaEndEffectorIndex, pos, orn, ll, ul,
        jr, rp, maxNumIterations=20)
      self.bullet_client.submitProfileTiming()
      for i in range(pandaNumDofs):
        self.bullet_client.setJointMotorControl2(self.panda, i, self.bullet_client.POSITION_CONTROL, jointPoses[i],force=5 * 240.)
      #target for fingers
    for i in [9,10]:
      self.bullet_client.setJointMotorControl2(self.panda, i, self.bullet_client.POSITION_CONTROL,self.finger_target ,force= 10)
    self.bullet_client.submitProfileTiming()


class PandaSimAuto(PandaSim):
  def __init__(self, bullet_client, offset):
    PandaSim.__init__(self, bullet_client, offset)
    self.state_t = 0
    self.cur_state = 0
    self.states=[0,3,5,4,6,3,7]
    self.state_durations=[1,1,1,2,1,1, 10]
  
  def update_state(self):
    self.state_t += self.control_dt
    if self.state_t > self.state_durations[self.cur_state]:
      self.cur_state += 1
      if self.cur_state>=len(self.states):
        self.cur_state = 0
      self.state_t = 0
      self.state=self.states[self.cur_state]
      #print("self.state=",self.state)



"""A wrapper that controls the timing between steps.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import time
import gin


@gin.configurable
class FixedSteptimeWrapperEnv(object):
  """A wrapped LocomotionGymEnv with timing control between steps."""

  def __init__(self,
               gym_env,
               desired_time_between_steps=None):
    """Initializes the wrapper env.

    Args:
      gym_env: An instance of LocomotionGymEnv.
      desired_time_between_steps: The desired time between steps in seconds.
        If this is None, it is set to the env_time_step of the gym_env.
    """
    self._gym_env = gym_env
    if desired_time_between_steps is None:
      self._desired_time_between_steps = gym_env.env_time_step
    else:
      self._desired_time_between_steps = desired_time_between_steps

    self._last_reset_time = time.time()
    self._last_step_time = time.time()
    self._step_counter = 0

  def __getattr__(self, attr):
    return getattr(self._gym_env, attr)

  def reset(self, initial_motor_angles=None, reset_duration=1.0):
    """Reset the environment.

    This function records the timing of the reset.

    Args:
      initial_motor_angles: Not used.
      reset_duration: Not used.

    Returns:
      The observation of the environment after reset.
    """
    obs = self._gym_env.reset(initial_motor_angles=initial_motor_angles,
                              reset_duration=reset_duration)
    self._last_reset_time = time.time()
    self._last_step_time = time.time()
    self._step_counter = 0
    return obs

  def step(self, action):
    """Steps the wrapped environment.

    Args:
      action: Numpy array. The input action from an NN agent.

    Returns:
      The tuple containing the observation, the reward, the epsiode end
      indicator.

    Raises:
      ValueError if input action is None.
    """
    time_between_steps = time.time() - self._last_step_time
    if time_between_steps < self._desired_time_between_steps:
      time.sleep(self._desired_time_between_steps - time_between_steps)
    self._last_step_time = time.time()
    self._step_counter += 1
    return self._gym_env.step(action)

  @property
  def elapsed_time(self):
    """Returns the elapsed time in seconds."""
    return time.time() - self._last_reset_time

  @property
  def steps_per_second(self):
    """Returns the average number of time steps for 1 second."""
    return self._step_counter / self.elapsed_time

  @property
  def seconds_per_step(self):
    """Returns the average time between steps."""
    return self.elapsed_time / self._step_counter



"""A gin-config class for locomotion_gym_env.

This should be identical to locomotion_gym_config.proto.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from typing import Sequence, Text

import attr
import gin


@gin.configurable
@attr.s
class SimulationParameters(object):
  """Parameters specific for the pyBullet simulation."""
  sim_time_step_s = attr.ib(type=float, default=0.002)
  num_action_repeat = attr.ib(type=int, default=5)
  enable_hard_reset = attr.ib(type=bool, default=False)
  enable_rendering = attr.ib(type=bool, default=False)
  enable_rendering_gui = attr.ib(type=bool, default=True)
  robot_on_rack = attr.ib(type=bool, default=False)
  camera_target = attr.ib(type=Sequence[float], default=None)
  camera_distance = attr.ib(type=float, default=1.0)
  camera_yaw = attr.ib(type=float, default=0)
  camera_pitch = attr.ib(type=float, default=-30)
  render_width = attr.ib(type=int, default=480)
  render_height = attr.ib(type=int, default=360)
  egl_rendering = attr.ib(type=bool, default=False)


@gin.configurable
@attr.s
class ScalarField(object):
  """A named scalar space with bounds."""
  # TODO(sehoonha) extension to vector fields.
  name = attr.ib(type=str)
  upper_bound = attr.ib(type=float)
  lower_bound = attr.ib(type=float)


@gin.configurable
@attr.s
class LocomotionGymConfig(object):
  """Grouped Config Parameters for LocomotionGym."""
  simulation_parameters = attr.ib(type=SimulationParameters)
  # TODO(sehoonha) implement attr validators for the list
  actions = attr.ib(type=list, default=None)  # pylint: disable=g-bare-generic
  log_path = attr.ib(type=Text, default=None)
  data_dir = attr.ib(
      type=Text,
      default='robotics/reinforcement_learning/minitaur/data/')
  profiling_path = attr.ib(type=Text, default=None)
  seed = attr.ib(type=int, default=None)
  ignored_sensor_list = attr.ib(type=Sequence[Text], default=())



# coding=utf-8
# Copyright 2020 The Google Research Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utility functions to calculate Minitaur's pose and motor angles."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import attr
import numpy as np

NUM_MOTORS = 8
NUM_LEGS = 4
MOTOR_SIGNS = (1, 1, -1, -1)
# Constants for the function swing_extend_to_motor_angles
EPS = 0.1
# Range of motion for the legs (does not allow pointing towards the body).
LEG_SWING_LIMIT_LOW = -math.pi / 2 + EPS
LEG_SWING_LIMIT_HIGH = 3 * math.pi / 2 - EPS
# Range of gap between motors for feasibility.
MOTORS_GAP_LIMIT_HIGH = 2 * math.pi - EPS
MOTORS_GAP_LIMIT_LOW = EPS


@attr.s
class MinitaurPose(object):
  """Default pose of the Minitaur."""
  swing_angle_0 = attr.ib(type=float, default=0)
  swing_angle_1 = attr.ib(type=float, default=0)
  swing_angle_2 = attr.ib(type=float, default=0)
  swing_angle_3 = attr.ib(type=float, default=0)
  extension_angle_0 = attr.ib(type=float, default=0)
  extension_angle_1 = attr.ib(type=float, default=0)
  extension_angle_2 = attr.ib(type=float, default=0)
  extension_angle_3 = attr.ib(type=float, default=0)


def motor_angles_to_leg_pose(motor_angles):
  """Convert motor angles to the leg pose.

  A single leg pose is a tuple (swing, extension). The definition can be find
  in:
    Sim-to-Real: Learning Agile Locomotion For Quadruped Robot

  Args:
    motor_angles: A numpy array. Contains all eight motor angles for Minitaur.

  Returns:
    A numpy array. Contains the leg pose for all four legs: [swing_0, swing_1,
    swing_2, swing_3, extension_0, extension_1, extension_2, extension_3]

  """
  motor_angles = np.array(motor_angles)

  swings = 0.5 * np.multiply(
      np.array(MOTOR_SIGNS), (motor_angles[1::2] - motor_angles[::2]))
  extensions = 0.5 * (motor_angles[::2] + motor_angles[1::2])

  return np.concatenate((swings, extensions), axis=None)


def leg_pose_to_motor_angles(leg_pose):
  """Converts the leg pose to the motor angles.

  Args:
    leg_pose: A numpy array. Contains the leg pose for all four legs: [swing_0,
      swing_1, swing_2, swing_3, extension_0, extension_1, extension_2,
      extension_3]

  Returns:
    A numpy array. All eight motor angles.
  """
  leg_pose = np.array(leg_pose)

  # swings multiplied with the sign array.
  signed_swings = np.multiply(np.array(MOTOR_SIGNS), leg_pose[0:NUM_LEGS])
  extensions = leg_pose[NUM_LEGS:]

  motor_angles = np.zeros(NUM_MOTORS)
  motor_angles[1::2] = signed_swings + extensions
  motor_angles[::2] = extensions - signed_swings
  return motor_angles


# This method also does the same conversion, but 0 swing and 0 extension maps
# to a neutral standing still motor positions with motors at + or - pi. It also
# contains a safety layer so that the legs don't swing or extend too much to hit
# the body of the robot.
def leg_pose_to_motor_angles_with_half_pi_offset_and_safety(leg_pose):
  """Converts the swing extension poses to the motor angles with safety limits.

  Args:
    leg_pose: A numpy array. Contains the leg pose for all four legs: [swing_0,
      extension_0, swing_1, extension_1, swing_2, extension_2, swing_3,
      extension_3]

  Returns:
    A numpy array. All eight motor angles.
  """

  motor_angles = []
  for idx in range(4):
    swing = leg_pose[idx * 2]
    extend = leg_pose[idx * 2 + 1]
    motor_angles.extend(swing_extend_to_motor_angles(idx, swing, extend))
  return motor_angles


def swing_extend_to_motor_angles(leg_id, swing, extension, noise_stdev=0):
  """Swing - extension based leg model for minitaur.

  Swing extension leg model calculates motor positions using 2 separate motions:
  swing and extension. Swing rotates the whole leg by rotating both motors
  equally towards same direction. Extension increases or decreases the length
  of the leg by turning both motors equally in opposite direction.

  This method also does the same conversion as leg_pose_to_motor_angles, but 0
  swing and 0 extension maps to a neutral standing still motor positions with
  motors at + or - pi.
  Args:
      leg_id: The id of the leg that the conversion is made for (0, 1, 2, 3).
      swing: Swing degree for the leg (in radians). 0 means perpendicular to the
        body).
      extension: Extension level (length) of the leg, limited to [-1, 1].
      noise_stdev: Standard deviation of the introduced noise at the motor
        position level. Noise is turned off by default.

  Returns:
    motor0: Position for the first motor for that leg.
    motor1: Position for the second motor for that leg.
  Raises:
    ValueError: In case calculated positions are outside the allowed boundaries.
  """
  # Check if the leg_id is in valid range
  if not 0 <= leg_id <= 3:
    raise ValueError('leg {} does not exist for a quadruped.'.format(leg_id))

  # Front legs can not swing too much towards the body.
  if leg_id % 2 == 0:
    swing = np.clip(swing, LEG_SWING_LIMIT_LOW, LEG_SWING_LIMIT_HIGH)
  # Back legs can not swing too much towards the body (opposite direction).
  else:
    swing = np.clip(swing, -LEG_SWING_LIMIT_HIGH, -LEG_SWING_LIMIT_LOW)

  # Check if the motors are too close or too far away to make it impossible
  # for the physical robot.
  gap = math.pi - 2 * extension
  if gap < MOTORS_GAP_LIMIT_LOW or gap > MOTORS_GAP_LIMIT_HIGH:
    top_extension = (math.pi - MOTORS_GAP_LIMIT_LOW) / 2.0
    least_extension = (math.pi - MOTORS_GAP_LIMIT_HIGH) / 2.0
    extension = np.clip(extension, least_extension, top_extension)

  # Initialization to neutral standing position where both motors point to
  # opposite directions
  motor0 = math.pi / 2
  motor1 = math.pi / 2
  # Rotational move
  if leg_id in (0, 1):
    motor0 += swing
    motor1 -= swing
  elif leg_id in (2, 3):
    motor0 -= swing
    motor1 += swing
  # Extension
  motor0 += extension
  motor1 += extension

  # Add noise if requested.
  if noise_stdev > 0:
    motor0 += np.random.normal(0, noise_stdev)
    motor1 += np.random.normal(0, noise_stdev)

  return motor0, motor1



# Copyright 2017 The TensorFlow Agents Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for the PPO algorithm."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import math
import re

import tf.compat.v1 as tf
from tensorflow.python.client import device_lib


def create_nested_vars(tensors):
  """Create variables matching a nested tuple of tensors.

  Args:
    tensors: Nested tuple of list of tensors.

  Returns:
    Nested tuple or list of variables.
  """
  if isinstance(tensors, (tuple, list)):
    return type(tensors)(create_nested_vars(tensor) for tensor in tensors)
  return tf.Variable(tensors, False)


def reinit_nested_vars(variables, indices=None):
  """Reset all variables in a nested tuple to zeros.

  Args:
    variables: Nested tuple or list of variaables.
    indices: Indices along the first dimension to reset, defaults to all.

  Returns:
    Operation.
  """
  if isinstance(variables, (tuple, list)):
    return tf.group(*[reinit_nested_vars(variable, indices) for variable in variables])
  if indices is None:
    return variables.assign(tf.zeros_like(variables))
  else:
    zeros = tf.zeros([tf.shape(indices)[0]] + variables.shape[1:].as_list())
    return tf.scatter_update(variables, indices, zeros)


def assign_nested_vars(variables, tensors):
  """Assign tensors to matching nested tuple of variables.

  Args:
    variables: Nested tuple or list of variables to update.
    tensors: Nested tuple or list of tensors to assign.

  Returns:
    Operation.
  """
  if isinstance(variables, (tuple, list)):
    return tf.group(
        *[assign_nested_vars(variable, tensor) for variable, tensor in zip(variables, tensors)])
  return variables.assign(tensors)


def discounted_return(reward, length, discount):
  """Discounted Monte-Carlo returns."""
  timestep = tf.range(reward.shape[1].value)
  mask = tf.cast(timestep[None, :] < length[:, None], tf.float32)
  return_ = tf.reverse(
      tf.transpose(
          tf.scan(lambda agg, cur: cur + discount * agg,
                  tf.transpose(tf.reverse(mask * reward, [1]), [1, 0]),
                  tf.zeros_like(reward[:, -1]), 1, False), [1, 0]), [1])
  return tf.check_numerics(tf.stop_gradient(return_), 'return')


def fixed_step_return(reward, value, length, discount, window):
  """N-step discounted return."""
  timestep = tf.range(reward.shape[1].value)
  mask = tf.cast(timestep[None, :] < length[:, None], tf.float32)
  return_ = tf.zeros_like(reward)
  for _ in range(window):
    return_ += reward
    reward = discount * tf.concat([reward[:, 1:], tf.zeros_like(reward[:, -1:])], 1)
  return_ += discount**window * tf.concat(
      [value[:, window:], tf.zeros_like(value[:, -window:]), 1])
  return tf.check_numerics(tf.stop_gradient(mask * return_), 'return')


def lambda_return(reward, value, length, discount, lambda_):
  """TD-lambda returns."""
  timestep = tf.range(reward.shape[1].value)
  mask = tf.cast(timestep[None, :] < length[:, None], tf.float32)
  sequence = mask * reward + discount * value * (1 - lambda_)
  discount = mask * discount * lambda_
  sequence = tf.stack([sequence, discount], 2)
  return_ = tf.reverse(
      tf.transpose(
          tf.scan(lambda agg, cur: cur[0] + cur[1] * agg,
                  tf.transpose(tf.reverse(sequence, [1]), [1, 2, 0]), tf.zeros_like(value[:, -1]),
                  1, False), [1, 0]), [1])
  return tf.check_numerics(tf.stop_gradient(return_), 'return')


def lambda_advantage(reward, value, length, discount):
  """Generalized Advantage Estimation."""
  timestep = tf.range(reward.shape[1].value)
  mask = tf.cast(timestep[None, :] < length[:, None], tf.float32)
  next_value = tf.concat([value[:, 1:], tf.zeros_like(value[:, -1:])], 1)
  delta = reward + discount * next_value - value
  advantage = tf.reverse(
      tf.transpose(
          tf.scan(lambda agg, cur: cur + discount * agg,
                  tf.transpose(tf.reverse(mask * delta, [1]), [1, 0]), tf.zeros_like(delta[:, -1]),
                  1, False), [1, 0]), [1])
  return tf.check_numerics(tf.stop_gradient(advantage), 'advantage')


def diag_normal_kl(mean0, logstd0, mean1, logstd1):
  """Epirical KL divergence of two normals with diagonal covariance."""
  logstd0_2, logstd1_2 = 2 * logstd0, 2 * logstd1
  return 0.5 * (tf.reduce_sum(tf.exp(logstd0_2 - logstd1_2), -1) + tf.reduce_sum(
      (mean1 - mean0)**2 / tf.exp(logstd1_2), -1) + tf.reduce_sum(logstd1_2, -1) -
                tf.reduce_sum(logstd0_2, -1) - mean0.shape[-1].value)


def diag_normal_logpdf(mean, logstd, loc):
  """Log density of a normal with diagonal covariance."""
  constant = -0.5 * (math.log(2 * math.pi) + logstd)
  value = -0.5 * ((loc - mean) / tf.exp(logstd))**2
  return tf.reduce_sum(constant + value, -1)


def diag_normal_entropy(mean, logstd):
  """Empirical entropy of a normal with diagonal covariance."""
  constant = mean.shape[-1].value * math.log(2 * math.pi * math.e)
  return (constant + tf.reduce_sum(2 * logstd, 1)) / 2


def available_gpus():
  """List of GPU device names detected by TensorFlow."""
  local_device_protos = device_lib.list_local_devices()
  return [x.name for x in local_device_protos if x.device_type == 'GPU']


def gradient_summaries(grad_vars, groups=None, scope='gradients'):
  """Create histogram summaries of the gradient.

  Summaries can be grouped via regexes matching variables names.

  Args:
    grad_vars: List of (gradient, variable) tuples as returned by optimizers.
    groups: Mapping of name to regex for grouping summaries.
    scope: Name scope for this operation.

  Returns:
    Summary tensor.
  """
  groups = groups or {r'all': r'.*'}
  grouped = collections.defaultdict(list)
  for grad, var in grad_vars:
    if grad is None:
      continue
    for name, pattern in groups.items():
      if re.match(pattern, var.name):
        name = re.sub(pattern, name, var.name)
        grouped[name].append(grad)
  for name in groups:
    if name not in grouped:
      tf.logging.warn("No variables matching '{}' group.".format(name))
  summaries = []
  for name, grads in grouped.items():
    grads = [tf.reshape(grad, [-1]) for grad in grads]
    grads = tf.concat(grads, 0)
    summaries.append(tf.summary.histogram(scope + '/' + name, grads))
  return tf.summary.merge(summaries)


def variable_summaries(vars_, groups=None, scope='weights'):
  """Create histogram summaries for the provided variables.

  Summaries can be grouped via regexes matching variables names.

  Args:
    vars_: List of variables to summarize.
    groups: Mapping of name to regex for grouping summaries.
    scope: Name scope for this operation.

  Returns:
    Summary tensor.
  """
  groups = groups or {r'all': r'.*'}
  grouped = collections.defaultdict(list)
  for var in vars_:
    for name, pattern in groups.items():
      if re.match(pattern, var.name):
        name = re.sub(pattern, name, var.name)
        grouped[name].append(var)
  for name in groups:
    if name not in grouped:
      tf.logging.warn("No variables matching '{}' group.".format(name))
  summaries = []
  for name, vars_ in grouped.items():
    vars_ = [tf.reshape(var, [-1]) for var in vars_]
    vars_ = tf.concat(vars_, 0)
    summaries.append(tf.summary.histogram(scope + '/' + name, vars_))
  return tf.summary.merge(summaries)



from pybullet_utils import pd_controller_stable
from pybullet_envs.deep_mimic.env import humanoid_pose_interpolator
import math
import numpy as np

chest = 1
neck = 2
rightHip = 3
rightKnee = 4
rightAnkle = 5
rightShoulder = 6
rightElbow = 7
leftHip = 9
leftKnee = 10
leftAnkle = 11
leftShoulder = 12
leftElbow = 13
jointFrictionForce = 0


class HumanoidStablePDMultiClip(object):

    def __init__(self, pybullet_client, mocap_data, timeStep,
                 useFixedBase=True, arg_parser=None, useComReward=False):
        self._pybullet_client = pybullet_client
        self._mocap_data = mocap_data # this is a dictionary
        self._arg_parser = arg_parser
        self._n_clips = self._mocap_data.getNumClips()
        print("LOADING humanoid!")
        flags = self._pybullet_client.URDF_MAINTAIN_LINK_ORDER + self._pybullet_client.URDF_USE_SELF_COLLISION + self._pybullet_client.URDF_USE_SELF_COLLISION_EXCLUDE_ALL_PARENTS
        self._sim_model = self._pybullet_client.loadURDF(
            "humanoid/humanoid.urdf", [0, 0.889540259, 0],
            globalScaling=0.25,
            useFixedBase=useFixedBase,
            flags=flags)

        self._end_effectors = [5, 8, 11, 14]  # ankle and wrist, both left and right

        self._kin_models = {}
        self._poseInterpolators = {}

        for i in range(self._n_clips):
            self._kin_models[i] = self._pybullet_client.loadURDF(
                "humanoid/humanoid.urdf", [0, 0.85, 0],
                globalScaling=0.25,
                useFixedBase=True,
                flags=self._pybullet_client.URDF_MAINTAIN_LINK_ORDER)

        self._pybullet_client.changeDynamics(self._sim_model, -1, lateralFriction=0.9)
        for j in range(self._pybullet_client.getNumJoints(self._sim_model)):
            self._pybullet_client.changeDynamics(self._sim_model, j, lateralFriction=0.9)

        self._pybullet_client.changeDynamics(self._sim_model, -1, linearDamping=0, angularDamping=0)
        for i in range(self._n_clips):
            self._pybullet_client.changeDynamics(self._kin_models[i], -1, linearDamping=0, angularDamping=0)

            # todo: add feature to disable simulation for a particular object. Until then, disable all collisions
            self._pybullet_client.setCollisionFilterGroupMask(self._kin_models[i],
                                                              -1,
                                                              collisionFilterGroup=0,
                                                              collisionFilterMask=0)
            self._pybullet_client.changeDynamics(
                self._kin_models[i],
                -1,
                activationState=self._pybullet_client.ACTIVATION_STATE_SLEEP +
                                self._pybullet_client.ACTIVATION_STATE_ENABLE_SLEEPING +
                                self._pybullet_client.ACTIVATION_STATE_DISABLE_WAKEUP)
            alpha = 0.4
            self._pybullet_client.changeVisualShape(self._kin_models[i], -1, rgbaColor=[1, 1, 1, alpha])
            for j in range(self._pybullet_client.getNumJoints(self._kin_models[i])):
                self._pybullet_client.setCollisionFilterGroupMask(self._kin_models[i],
                                                                  j,
                                                                  collisionFilterGroup=0,
                                                                  collisionFilterMask=0)
                self._pybullet_client.changeDynamics(
                    self._kin_models[i],
                    j,
                    activationState=self._pybullet_client.ACTIVATION_STATE_SLEEP +
                                    self._pybullet_client.ACTIVATION_STATE_ENABLE_SLEEPING +
                                    self._pybullet_client.ACTIVATION_STATE_DISABLE_WAKEUP)
                self._pybullet_client.changeVisualShape(self._kin_models[i], j, rgbaColor=[1, 1, 1, alpha])

            self._poseInterpolators[i] = humanoid_pose_interpolator.HumanoidPoseInterpolator()

            for j in range(self._mocap_data.getNumFrames() - 1):
                frameData = self._mocap_data._motion_data[i]['Frames'][j]
                self._poseInterpolators[i].PostProcessMotionData(frameData)

        self._stablePD = pd_controller_stable.PDControllerStableMultiDof(self._pybullet_client)
        self._timeStep = timeStep
        self._kpOrg = [
            0, 0, 0, 0, 0, 0, 0, 1000, 1000, 1000, 1000, 100, 100, 100, 100, 500, 500, 500, 500, 500,
            400, 400, 400, 400, 400, 400, 400, 400, 300, 500, 500, 500, 500, 500, 400, 400, 400, 400,
            400, 400, 400, 400, 300
        ]
        self._kdOrg = [
            0, 0, 0, 0, 0, 0, 0, 100, 100, 100, 100, 10, 10, 10, 10, 50, 50, 50, 50, 50, 40, 40, 40,
            40, 40, 40, 40, 40, 30, 50, 50, 50, 50, 50, 40, 40, 40, 40, 40, 40, 40, 40, 30
        ]

        self._jointIndicesAll = [
            chest, neck, rightHip, rightKnee, rightAnkle, rightShoulder, rightElbow, leftHip, leftKnee,
            leftAnkle, leftShoulder, leftElbow
        ]
        for j in self._jointIndicesAll:
            self._pybullet_client.setJointMotorControl2(self._sim_model,
                                                        j,
                                                        self._pybullet_client.POSITION_CONTROL,
                                                        targetPosition=0,
                                                        positionGain=0,
                                                        targetVelocity=0,
                                                        force=jointFrictionForce)
            self._pybullet_client.setJointMotorControlMultiDof(
                self._sim_model,
                j,
                self._pybullet_client.POSITION_CONTROL,
                targetPosition=[0, 0, 0, 1],
                targetVelocity=[0, 0, 0],
                positionGain=0,
                velocityGain=1,
                force=[jointFrictionForce, jointFrictionForce, jointFrictionForce])

            for i in range(self._n_clips):
                self._pybullet_client.setJointMotorControl2(self._kin_models[i],
                                                            j,
                                                            self._pybullet_client.POSITION_CONTROL,
                                                            targetPosition=0,
                                                            positionGain=0,
                                                            targetVelocity=0,
                                                            force=0)
                self._pybullet_client.setJointMotorControlMultiDof(
                    self._kin_models[i],
                    j,
                    self._pybullet_client.POSITION_CONTROL,
                    targetPosition=[0, 0, 0, 1],
                    targetVelocity=[0, 0, 0],
                    positionGain=0,
                    velocityGain=1,
                    force=[jointFrictionForce, jointFrictionForce, 0])

        self._jointDofCounts = [4, 4, 4, 1, 4, 4, 1, 4, 1, 4, 4, 1]

        # only those body parts/links are allowed to touch the ground, otherwise the episode terminates
        fall_contact_bodies = []
        if self._arg_parser is not None:
            fall_contact_bodies = self._arg_parser.parse_ints("fall_contact_bodies")
        self._fall_contact_body_parts = fall_contact_bodies

        # [x,y,z] base position and [x,y,z,w] base orientation
        self._totalDofs = 7
        for dof in self._jointDofCounts:
            self._totalDofs += dof
        self.setSimTime(0)

        self._useComReward = useComReward

        self.resetPose()

    def resetPose(self):
        # print("resetPose with self._frame=", self._frame, " and self._frameFraction=",self._frameFraction)
        pose = self.computePose(self._frameFraction, 0)
        self.initializePose(self._poseInterpolators[0], self._sim_model, initBase=True)
        self.initializePose(self._poseInterpolators[0], self._kin_models[0], initBase=False)
        for i in range(1, self._n_clips):
            _ = self.computePose(self._frameFraction, i)
            self.initializePose(self._poseInterpolators[i], self._kin_models[i], initBase=False)

    def initializePose(self, pose, phys_model, initBase, initializeVelocity=True):

        if initializeVelocity:
            if initBase:
                self._pybullet_client.resetBasePositionAndOrientation(phys_model, pose._basePos,
                                                                      pose._baseOrn)
                self._pybullet_client.resetBaseVelocity(phys_model, pose._baseLinVel, pose._baseAngVel)

            indices = [chest, neck, rightHip, rightKnee,
                       rightAnkle, rightShoulder, rightElbow, leftHip,
                       leftKnee, leftAnkle, leftShoulder, leftElbow]
            jointPositions = [pose._chestRot, pose._neckRot, pose._rightHipRot, pose._rightKneeRot,
                              pose._rightAnkleRot, pose._rightShoulderRot, pose._rightElbowRot, pose._leftHipRot,
                              pose._leftKneeRot, pose._leftAnkleRot, pose._leftShoulderRot, pose._leftElbowRot]

            jointVelocities = [pose._chestVel, pose._neckVel, pose._rightHipVel, pose._rightKneeVel,
                               pose._rightAnkleVel, pose._rightShoulderVel, pose._rightElbowVel, pose._leftHipVel,
                               pose._leftKneeVel, pose._leftAnkleVel, pose._leftShoulderVel, pose._leftElbowVel]
            self._pybullet_client.resetJointStatesMultiDof(phys_model, indices,
                                                           jointPositions, jointVelocities)
        else:
            if initBase:
                self._pybullet_client.resetBasePositionAndOrientation(phys_model, pose._basePos,
                                                                      pose._baseOrn)

            indices = [chest, neck, rightHip, rightKnee,
                       rightAnkle, rightShoulder, rightElbow, leftHip,
                       leftKnee, leftAnkle, leftShoulder, leftElbow]
            jointPositions = [pose._chestRot, pose._neckRot, pose._rightHipRot, pose._rightKneeRot,
                              pose._rightAnkleRot, pose._rightShoulderRot, pose._rightElbowRot, pose._leftHipRot,
                              pose._leftKneeRot, pose._leftAnkleRot, pose._leftShoulderRot, pose._leftElbowRot]
            self._pybullet_client.resetJointStatesMultiDof(phys_model, indices, jointPositions)

    def calcCycleCount(self, simTime, cycleTime):
        phases = simTime / cycleTime
        count = math.floor(phases)
        return count

    def getCycleTime(self):
        keyFrameDuration = self._mocap_data.getKeyFrameDuration()
        cycleTime = keyFrameDuration * (self._mocap_data.getNumFrames() - 1)
        return cycleTime

    def setSimTime(self, t):
        self._simTime = t
        # print("SetTimeTime time =",t)
        keyFrameDuration = self._mocap_data.getKeyFrameDuration()
        cycleTime = self.getCycleTime()
        # print("self._motion_data.NumFrames()=",self._mocap_data.NumFrames())
        self._cycleCount = self.calcCycleCount(t, cycleTime)
        # print("cycles=",cycles)
        frameTime = t - self._cycleCount * cycleTime
        if (frameTime < 0):
            frameTime += cycleTime

        # print("keyFrameDuration=",keyFrameDuration)
        # print("frameTime=",frameTime)
        self._frame = int(frameTime / keyFrameDuration)
        # print("self._frame=",self._frame)

        self._frameNext = self._frame + 1
        if (self._frameNext >= self._mocap_data.getNumFrames()):
            self._frameNext = self._frame

        self._frameFraction = (frameTime - self._frame * keyFrameDuration) / (keyFrameDuration)

    def computeCycleOffset(self, i=0):
        lastFrame = self._mocap_data.getNumFrames() - 1
        frameData = self._mocap_data._motion_data[i]['Frames'][0]
        frameDataNext = self._mocap_data._motion_data[i]['Frames'][lastFrame]

        basePosStart = [frameData[1], frameData[2], frameData[3]]
        basePosEnd = [frameDataNext[1], frameDataNext[2], frameDataNext[3]]
        self._cycleOffset = [
            basePosEnd[0] - basePosStart[0], basePosEnd[1] - basePosStart[1],
            basePosEnd[2] - basePosStart[2]
        ]
        return self._cycleOffset

    def computePose(self, frameFraction, i=0):
        frameData = self._mocap_data._motion_data[i]['Frames'][self._frame]
        frameDataNext = self._mocap_data._motion_data[i]['Frames'][self._frameNext]

        self._poseInterpolators[i].Slerp(frameFraction, frameData, frameDataNext, self._pybullet_client)
        # print("self._poseInterpolator.Slerp(", frameFraction,")=", pose)
        self.computeCycleOffset()
        oldPos = self._poseInterpolators[i]._basePos
        self._poseInterpolators[i]._basePos = [
            oldPos[0] + self._cycleCount * self._cycleOffset[0],
            oldPos[1] + self._cycleCount * self._cycleOffset[1],
            oldPos[2] + self._cycleCount * self._cycleOffset[2]
        ]
        pose = self._poseInterpolators[i].GetPose()

        return pose

    def convertActionToPose(self, action, i):
        pose = self._poseInterpolators[i].ConvertFromAction(self._pybullet_client, action)
        return pose

    def computeAndApplyPDForces(self, desiredPositions, maxForces):
        dofIndex = 7
        scaling = 1
        indices = []
        forces = []
        targetPositions = []
        targetVelocities = []
        kps = []
        kds = []

        for index in range(len(self._jointIndicesAll)):
            jointIndex = self._jointIndicesAll[index]
            indices.append(jointIndex)
            kps.append(self._kpOrg[dofIndex])
            kds.append(self._kdOrg[dofIndex])
            if self._jointDofCounts[index] == 4:
                force = [
                    scaling * maxForces[dofIndex + 0],
                    scaling * maxForces[dofIndex + 1],
                    scaling * maxForces[dofIndex + 2]
                ]
                targetVelocity = [0, 0, 0]
                targetPosition = [
                    desiredPositions[dofIndex + 0],
                    desiredPositions[dofIndex + 1],
                    desiredPositions[dofIndex + 2],
                    desiredPositions[dofIndex + 3]
                ]
            if self._jointDofCounts[index] == 1:
                force = [scaling * maxForces[dofIndex]]
                targetPosition = [desiredPositions[dofIndex + 0]]
                targetVelocity = [0]
            forces.append(force)
            targetPositions.append(targetPosition)
            targetVelocities.append(targetVelocity)
            dofIndex += self._jointDofCounts[index]


        self._pybullet_client.setJointMotorControlMultiDofArray(self._sim_model,
                                                                indices,
                                                                self._pybullet_client.STABLE_PD_CONTROL,
                                                                targetPositions=targetPositions,
                                                                targetVelocities=targetVelocities,
                                                                forces=forces,
                                                                positionGains=kps,
                                                                velocityGains=kds,
                                                                )


    def getPhase(self):
        keyFrameDuration = self._mocap_data.getKeyFrameDuration()
        cycleTime = keyFrameDuration * (self._mocap_data.getNumFrames() - 1)
        phase = self._simTime / cycleTime
        phase = math.fmod(phase, 1.0)
        if (phase < 0):
            phase += 1
        return phase

    def buildHeadingTrans(self, rootOrn):
        # align root transform 'forward' with world-space x axis
        eul = self._pybullet_client.getEulerFromQuaternion(rootOrn)
        refDir = [1, 0, 0]
        rotVec = self._pybullet_client.rotateVector(rootOrn, refDir)
        heading = math.atan2(-rotVec[2], rotVec[0])
        heading2 = eul[1]
        # print("heading=",heading)
        headingOrn = self._pybullet_client.getQuaternionFromAxisAngle([0, 1, 0], -heading)
        return headingOrn

    def buildOriginTrans(self):
        rootPos, rootOrn = self._pybullet_client.getBasePositionAndOrientation(self._sim_model)

        # print("rootPos=",rootPos, " rootOrn=",rootOrn)
        invRootPos = [-rootPos[0], 0, -rootPos[2]]
        # invOrigTransPos, invOrigTransOrn = self._pybullet_client.invertTransform(rootPos,rootOrn)
        headingOrn = self.buildHeadingTrans(rootOrn)
        # print("headingOrn=",headingOrn)
        headingMat = self._pybullet_client.getMatrixFromQuaternion(headingOrn)
        # print("headingMat=",headingMat)
        # dummy, rootOrnWithoutHeading = self._pybullet_client.multiplyTransforms([0,0,0],headingOrn, [0,0,0], rootOrn)
        # dummy, invOrigTransOrn = self._pybullet_client.multiplyTransforms([0,0,0],rootOrnWithoutHeading, invOrigTransPos, invOrigTransOrn)

        invOrigTransPos, invOrigTransOrn = self._pybullet_client.multiplyTransforms([0, 0, 0],
                                                                                    headingOrn,
                                                                                    invRootPos,
                                                                                    [0, 0, 0, 1])
        # print("invOrigTransPos=",invOrigTransPos)
        # print("invOrigTransOrn=",invOrigTransOrn)
        invOrigTransMat = self._pybullet_client.getMatrixFromQuaternion(invOrigTransOrn)
        # print("invOrigTransMat =",invOrigTransMat )
        return invOrigTransPos, invOrigTransOrn

    def getState(self):

        stateVector = []
        phase = self.getPhase()
        # print("phase=",phase)
        stateVector.append(phase)

        rootTransPos, rootTransOrn = self.buildOriginTrans()
        basePos, baseOrn = self._pybullet_client.getBasePositionAndOrientation(self._sim_model)

        rootPosRel, dummy = self._pybullet_client.multiplyTransforms(rootTransPos, rootTransOrn,
                                                                     basePos, [0, 0, 0, 1])
        # print("!!!rootPosRel =",rootPosRel )
        # print("rootTransPos=",rootTransPos)
        # print("basePos=",basePos)
        localPos, localOrn = self._pybullet_client.multiplyTransforms(rootTransPos, rootTransOrn,
                                                                      basePos, baseOrn)

        localPos = [
            localPos[0] - rootPosRel[0], localPos[1] - rootPosRel[1], localPos[2] - rootPosRel[2]
        ]
        # print("localPos=",localPos)

        stateVector.append(rootPosRel[1])

        self.pb2dmJoints = range(15)

        linkIndicesSim = []
        for pbJoint in range(self._pybullet_client.getNumJoints(self._sim_model)):
            linkIndicesSim.append(self.pb2dmJoints[pbJoint])

        linkStatesSim = self._pybullet_client.getLinkStates(self._sim_model, linkIndicesSim,
                                                            computeForwardKinematics=True, computeLinkVelocity=True)

        for pbJoint in range(self._pybullet_client.getNumJoints(self._sim_model)):
            j = self.pb2dmJoints[pbJoint]
            # print("joint order:",j)
            ls = linkStatesSim[pbJoint]
            linkPos = ls[0]
            linkOrn = ls[1]
            linkPosLocal, linkOrnLocal = self._pybullet_client.multiplyTransforms(
                rootTransPos, rootTransOrn, linkPos, linkOrn)
            if (linkOrnLocal[3] < 0):
                linkOrnLocal = [-linkOrnLocal[0], -linkOrnLocal[1], -linkOrnLocal[2], -linkOrnLocal[3]]
            linkPosLocal = [
                linkPosLocal[0] - rootPosRel[0], linkPosLocal[1] - rootPosRel[1],
                linkPosLocal[2] - rootPosRel[2]
            ]
            for l in linkPosLocal:
                stateVector.append(l)
            # re-order the quaternion, DeepMimic uses w,x,y,z

            if (linkOrnLocal[3] < 0):
                linkOrnLocal[0] *= -1
                linkOrnLocal[1] *= -1
                linkOrnLocal[2] *= -1
                linkOrnLocal[3] *= -1

            stateVector.append(linkOrnLocal[3])
            stateVector.append(linkOrnLocal[0])
            stateVector.append(linkOrnLocal[1])
            stateVector.append(linkOrnLocal[2])

        for pbJoint in range(self._pybullet_client.getNumJoints(self._sim_model)):
            j = self.pb2dmJoints[pbJoint]
            ls = linkStatesSim[pbJoint]

            linkLinVel = ls[6]
            linkAngVel = ls[7]
            linkLinVelLocal, unused = self._pybullet_client.multiplyTransforms([0, 0, 0], rootTransOrn,
                                                                               linkLinVel, [0, 0, 0, 1])

            linkAngVelLocal, unused = self._pybullet_client.multiplyTransforms([0, 0, 0], rootTransOrn,
                                                                               linkAngVel, [0, 0, 0, 1])

            for l in linkLinVelLocal:
                stateVector.append(l)
            for l in linkAngVelLocal:
                stateVector.append(l)

        # print("stateVector len=",len(stateVector))
        # for st in range (len(stateVector)):
        #  print("state[",st,"]=",stateVector[st])
        return stateVector

    def terminates(self):
        # check if any non-allowed body part hits the ground
        terminates = False
        pts = self._pybullet_client.getContactPoints()
        for p in pts:
            part = -1
            # ignore self-collision
            if (p[1] == p[2]):
                continue
            if (p[1] == self._sim_model):
                part = p[3]
            if (p[2] == self._sim_model):
                part = p[4]
            if (part >= 0 and part in self._fall_contact_body_parts):
                # print("terminating part:", part)
                terminates = True

        return terminates

    def quatMul(self, q1, q2):
        return [
            q1[3] * q2[0] + q1[0] * q2[3] + q1[1] * q2[2] - q1[2] * q2[1],
            q1[3] * q2[1] + q1[1] * q2[3] + q1[2] * q2[0] - q1[0] * q2[2],
            q1[3] * q2[2] + q1[2] * q2[3] + q1[0] * q2[1] - q1[1] * q2[0],
            q1[3] * q2[3] - q1[0] * q2[0] - q1[1] * q2[1] - q1[2] * q2[2]
        ]

    def calcRootAngVelErr(self, vel0, vel1):
        diff = [vel0[0] - vel1[0], vel0[1] - vel1[1], vel0[2] - vel1[2]]
        return diff[0] * diff[0] + diff[1] * diff[1] + diff[2] * diff[2]

    def calcRootRotDiff(self, orn0, orn1):
        orn0Conj = [-orn0[0], -orn0[1], -orn0[2], orn0[3]]
        q_diff = self.quatMul(orn1, orn0Conj)
        axis, angle = self._pybullet_client.getAxisAngleFromQuaternion(q_diff)
        return angle * angle

    def getReward(self, pose, i=0):
        """Compute and return the pose-based reward."""
        # from DeepMimic double cSceneImitate::CalcRewardImitate
        # todo: compensate for ground height in some parts, once we move to non-flat terrain
        # not values from the paper, but from the published code.
        pose_w = 0.5
        vel_w = 0.05
        end_eff_w = 0.15
        # does not exist in paper
        root_w = 0.2
        if self._useComReward:
            com_w = 0.1
        else:
            com_w = 0

        total_w = pose_w + vel_w + end_eff_w + root_w + com_w
        pose_w /= total_w
        vel_w /= total_w
        end_eff_w /= total_w
        root_w /= total_w
        com_w /= total_w

        pose_scale = 2
        vel_scale = 0.1
        end_eff_scale = 40
        root_scale = 5
        com_scale = 10
        err_scale = 1  # error scale

        reward = 0

        pose_err = 0
        vel_err = 0
        end_eff_err = 0
        root_err = 0
        com_err = 0
        heading_err = 0

        # create a mimic reward, comparing the dynamics humanoid with a kinematic one

        if self._useComReward:
            comSim, comSimVel = self.computeCOMposVel(self._sim_model)
            comKin, comKinVel = self.computeCOMposVel(self._kin_models[i])

        root_id = 0

        mJointWeights = [
            0.20833, 0.10416, 0.0625, 0.10416, 0.0625, 0.041666666666666671, 0.0625, 0.0416, 0.00,
            0.10416, 0.0625, 0.0416, 0.0625, 0.0416, 0.0000
        ]

        num_end_effs = 0
        num_joints = 15

        root_rot_w = mJointWeights[root_id]
        rootPosSim, rootOrnSim = self._pybullet_client.getBasePositionAndOrientation(self._sim_model)
        rootPosKin, rootOrnKin = self._pybullet_client.getBasePositionAndOrientation(self._kin_models[i])
        linVelSim, angVelSim = self._pybullet_client.getBaseVelocity(self._sim_model)
        # don't read the velocities from the kinematic model (they are zero), use the pose interpolator velocity
        # see also issue https://github.com/bulletphysics/bullet3/issues/2401
        linVelKin = self._poseInterpolators[i]._baseLinVel
        angVelKin = self._poseInterpolators[i]._baseAngVel

        root_rot_err = self.calcRootRotDiff(rootOrnSim, rootOrnKin)
        pose_err += root_rot_w * root_rot_err

        root_vel_diff = [
            linVelSim[0] - linVelKin[0], linVelSim[1] - linVelKin[1], linVelSim[2] - linVelKin[2]
        ]
        root_vel_err = root_vel_diff[0] * root_vel_diff[0] + root_vel_diff[1] * root_vel_diff[
            1] + root_vel_diff[2] * root_vel_diff[2]

        root_ang_vel_err = self.calcRootAngVelErr(angVelSim, angVelKin)
        vel_err += root_rot_w * root_ang_vel_err

        jointIndices = range(num_joints)
        simJointStates = self._pybullet_client.getJointStatesMultiDof(self._sim_model, jointIndices)
        kinJointStates = self._pybullet_client.getJointStatesMultiDof(self._kin_models[i], jointIndices)
        linkStatesSim = self._pybullet_client.getLinkStates(self._sim_model, jointIndices)
        linkStatesKin = self._pybullet_client.getLinkStates(self._kin_models[i], jointIndices)

        for j in range(num_joints):
            curr_pose_err = 0
            curr_vel_err = 0
            w = mJointWeights[j]

            simJointInfo = simJointStates[j]

            # print("simJointInfo.pos=",simJointInfo[0])
            # print("simJointInfo.vel=",simJointInfo[1])

            kinJointInfo = kinJointStates[j]

            # print("kinJointInfo.pos=",kinJointInfo[0])
            # print("kinJointInfo.vel=",kinJointInfo[1])
            if (len(simJointInfo[0]) == 1):
                angle = simJointInfo[0][0] - kinJointInfo[0][0]
                curr_pose_err = angle * angle
                velDiff = simJointInfo[1][0] - kinJointInfo[1][0]
                curr_vel_err = velDiff * velDiff
            if (len(simJointInfo[0]) == 4):
                # print("quaternion diff")
                diffQuat = self._pybullet_client.getDifferenceQuaternion(simJointInfo[0], kinJointInfo[0])
                axis, angle = self._pybullet_client.getAxisAngleFromQuaternion(diffQuat)
                curr_pose_err = angle * angle
                diffVel = [
                    simJointInfo[1][0] - kinJointInfo[1][0], simJointInfo[1][1] - kinJointInfo[1][1],
                    simJointInfo[1][2] - kinJointInfo[1][2]
                ]
                curr_vel_err = diffVel[0] * diffVel[0] + diffVel[1] * diffVel[1] + diffVel[2] * diffVel[2]

            pose_err += w * curr_pose_err
            vel_err += w * curr_vel_err

            is_end_eff = j in self._end_effectors

            if is_end_eff:

                linkStateSim = linkStatesSim[j]
                linkStateKin = linkStatesKin[j]

                linkPosSim = linkStateSim[0]
                linkPosKin = linkStateKin[0]
                linkPosDiff = [
                    linkPosSim[0] - linkPosKin[0], linkPosSim[1] - linkPosKin[1],
                    linkPosSim[2] - linkPosKin[2]
                ]
                curr_end_err = linkPosDiff[0] * linkPosDiff[0] + linkPosDiff[1] * linkPosDiff[
                    1] + linkPosDiff[2] * linkPosDiff[2]
                end_eff_err += curr_end_err
                num_end_effs += 1

        if (num_end_effs > 0):
            end_eff_err /= num_end_effs

        root_pos_diff = [
            rootPosSim[0] - rootPosKin[0], rootPosSim[1] - rootPosKin[1], rootPosSim[2] - rootPosKin[2]
        ]
        root_pos_err = root_pos_diff[0] * root_pos_diff[0] + root_pos_diff[1] * root_pos_diff[
            1] + root_pos_diff[2] * root_pos_diff[2]

        root_err = root_pos_err + 0.1 * root_rot_err + 0.01 * root_vel_err + 0.001 * root_ang_vel_err

        # COM error in initial code -> COM velocities
        if self._useComReward:
            com_err = 0.1 * np.sum(np.square(comKinVel - comSimVel))
        # com_err = 0.1 * np.sum(np.square(comKin - comSim))
        # com_err = 0.1 * (com_vel1_world - com_vel0_world).squaredNorm()

        # print("pose_err=",pose_err)
        # print("vel_err=",vel_err)
        pose_reward = math.exp(-err_scale * pose_scale * pose_err)
        vel_reward = math.exp(-err_scale * vel_scale * vel_err)
        end_eff_reward = math.exp(-err_scale * end_eff_scale * end_eff_err)
        root_reward = math.exp(-err_scale * root_scale * root_err)
        com_reward = math.exp(-err_scale * com_scale * com_err)

        reward = pose_w * pose_reward + vel_w * vel_reward + end_eff_w * end_eff_reward + root_w * root_reward + com_w * com_reward

        # pose_reward,vel_reward,end_eff_reward, root_reward, com_reward);
        # print("reward=",reward)
        # print("pose_reward=",pose_reward)
        # print("vel_reward=",vel_reward)
        # print("end_eff_reward=",end_eff_reward)
        # print("root_reward=",root_reward)
        # print("com_reward=",com_reward)

        return reward

    def computeCOMposVel(self, uid: int):
        """Compute center-of-mass position and velocity."""
        pb = self._pybullet_client
        num_joints = 15
        jointIndices = range(num_joints)
        link_states = pb.getLinkStates(uid, jointIndices, computeLinkVelocity=1)
        link_pos = np.array([s[0] for s in link_states])
        link_vel = np.array([s[-2] for s in link_states])
        tot_mass = 0.
        masses = []
        for j in jointIndices:
            mass_, *_ = pb.getDynamicsInfo(uid, j)
            masses.append(mass_)
            tot_mass += mass_
        masses = np.asarray(masses)[:, None]
        com_pos = np.sum(masses * link_pos, axis=0) / tot_mass
        com_vel = np.sum(masses * link_vel, axis=0) / tot_mass
        return com_pos, com_vel



import pybullet as p
from time import sleep

physicsClient = p.connect(p.GUI)
import pybullet_data

p.setAdditionalSearchPath(pybullet_data.getDataPath())
p.resetSimulation(p.RESET_USE_DEFORMABLE_WORLD)

gravZ=-10
p.setGravity(0, 0, gravZ)

planeOrn = [0,0,0,1]#p.getQuaternionFromEuler([0.3,0,0])
#planeId = p.loadURDF("plane.urdf", [0,0,-2],planeOrn)

boxId = p.loadURDF("cube.urdf", [0,1,2],useMaximalCoordinates = True)

clothId = p.loadSoftBody("cloth_z_up.obj", basePosition = [0,0,2], scale = 0.5, mass = 1., useNeoHookean = 0, useBendingSprings=1,useMassSpring=1, springElasticStiffness=40, springDampingStiffness=.1, springDampingAllDirections = 1, useSelfCollision = 0, frictionCoeff = .5, useFaceContact=1)


p.changeVisualShape(clothId, -1, flags=p.VISUAL_SHAPE_DOUBLE_SIDED)

p.createSoftBodyAnchor(clothId  ,24,-1,-1)
p.createSoftBodyAnchor(clothId ,20,-1,-1)
p.createSoftBodyAnchor(clothId ,15,boxId,-1, [0.5,-0.5,0])
p.createSoftBodyAnchor(clothId ,19,boxId,-1, [-0.5,-0.5,0])
p.setPhysicsEngineParameter(sparseSdfVoxelSize=0.25)

debug = True
if debug:
  data = p.getMeshData(clothId, -1, flags=p.MESH_DATA_SIMULATION_MESH)
  print("--------------")
  print("data=",data)
  print(data[0])
  print(data[1])
  text_uid = []
  for i in range(data[0]):
      pos = data[1][i]
      uid = p.addUserDebugText(str(i), pos, textColorRGB=[1,1,1])
      text_uid.append(uid)

while p.isConnected():
  p.getCameraImage(320,200)
  
  if debug:
    data = p.getMeshData(clothId, -1, flags=p.MESH_DATA_SIMULATION_MESH)
    for i in range(data[0]):
      pos = data[1][i]
      uid = p.addUserDebugText(str(i), pos, textColorRGB=[1,1,1], replaceItemUniqueId=text_uid[i])

  p.setGravity(0,0,gravZ)
  p.stepSimulation()
  #p.configureDebugVisualizer(p.COV_ENABLE_SINGLE_STEP_RENDERING,1)
  #sleep(1./240.)
  



import numpy as np
import math
from pybullet_envs.deep_mimic.env.env import Env
from pybullet_envs.deep_mimic.env.action_space import ActionSpace
from pybullet_utils import bullet_client
import time
from pybullet_envs.deep_mimic.env import motion_capture_data_multiclip
from pybullet_envs.deep_mimic.env import humanoid_stable_pd_multiclip
import pybullet_data
import pybullet as p1
import random

from enum import Enum


class InitializationStrategy(Enum):
    """Set how the environment is initialized."""
    START = 0
    RANDOM = 1  # random state initialization (RSI)


class PyBulletDeepMimicEnvMultiClip(Env):

    def __init__(self, arg_parser=None, enable_draw=False, pybullet_client=None,
                 time_step=1. / 240,
                 init_strategy=InitializationStrategy.RANDOM):
        super().__init__(arg_parser, enable_draw)
        self._num_agents = 1
        self._pybullet_client = pybullet_client
        self._isInitialized = False
        #self._useStablePD = True
        self._arg_parser = arg_parser
        self.timeStep = time_step
        self._init_strategy = init_strategy
        self._n_clips = -1
        print("Initialization strategy: {:s}".format(init_strategy))
        self.reset()

    def reset(self):

        if not self._isInitialized:
            if self.enable_draw:
                self._pybullet_client = bullet_client.BulletClient(connection_mode=p1.GUI)
                # disable 'GUI' since it slows down a lot on Mac OSX and some other platforms
                self._pybullet_client.configureDebugVisualizer(self._pybullet_client.COV_ENABLE_GUI, 0)
            else:
                self._pybullet_client = bullet_client.BulletClient()

            self._pybullet_client.setAdditionalSearchPath(pybullet_data.getDataPath())
            z2y = self._pybullet_client.getQuaternionFromEuler([-math.pi * 0.5, 0, 0])
            self._planeId = self._pybullet_client.loadURDF("plane_implicit.urdf", [0, 0, 0],
                                                           z2y,
                                                           useMaximalCoordinates=True)
            # print("planeId=",self._planeId)
            self._pybullet_client.configureDebugVisualizer(self._pybullet_client.COV_ENABLE_Y_AXIS_UP, 1)
            self._pybullet_client.setGravity(0, -9.8, 0)

            self._pybullet_client.setPhysicsEngineParameter(numSolverIterations=10)
            self._pybullet_client.changeDynamics(self._planeId, linkIndex=-1, lateralFriction=0.9)

            self._mocapData = motion_capture_data_multiclip.MotionCaptureDataMultiClip()

            motion_file = self._arg_parser.parse_strings('motion_file')
            print("motion_file=", motion_file[0])

            motionPath = pybullet_data.getDataPath() + "/" + motion_file[0]

            self._mocapData.Load(motionPath)
            self._n_clips = self._mocapData.getNumClips()
            timeStep = self.timeStep
            useFixedBase = False
            self._humanoid = humanoid_stable_pd_multiclip.HumanoidStablePDMultiClip(self._pybullet_client, self._mocapData,
                                                                                    timeStep, useFixedBase, self._arg_parser)
            self._isInitialized = True

            self._pybullet_client.setTimeStep(timeStep)
            self._pybullet_client.setPhysicsEngineParameter(numSubSteps=1)

        # print("numframes = ", self._humanoid._mocap_data.NumFrames())

        if self._init_strategy == InitializationStrategy.RANDOM:
            rnrange = 1000
            rn = random.randint(0, rnrange)
            startTime = float(rn) / rnrange * self._humanoid.getCycleTime()
        elif self._init_strategy == InitializationStrategy.START:
            startTime = 0

        self.t = startTime

        self._humanoid.setSimTime(startTime)

        self._humanoid.resetPose()
        # this clears the contact points. Todo: add API to explicitly clear all contact points?
        self._humanoid.resetPose()
        self.needs_update_time = self.t - 1  # force update

    def get_num_agents(self):
        return self._num_agents

    def get_action_space(self, agent_id):
        return ActionSpace(ActionSpace.Continuous)

    def get_reward_min(self, agent_id):
        return 0

    def get_reward_max(self, agent_id):
        return 1

    def get_reward_fail(self, agent_id):
        return self.get_reward_min(agent_id)

    def get_reward_succ(self, agent_id):
        return self.get_reward_max(agent_id)

    # scene_name == "imitate" -> cDrawSceneImitate
    def get_state_size(self, agent_id):
        # cCtController::GetStateSize()
        # int state_size = cDeepMimicCharController::GetStateSize();
        #                     state_size += GetStatePoseSize();#106
        #                     state_size += GetStateVelSize(); #(3+3)*numBodyParts=90
        # state_size += GetStatePhaseSize();#1
        # 197
        return 197

    def build_state_norm_groups(self, agent_id):
        groups = [0] * self.get_state_size(agent_id)
        groups[0] = -1
        return groups

    def build_state_offset(self, agent_id):
        out_offset = [0] * self.get_state_size(agent_id)
        phase_offset = -0.5
        out_offset[0] = phase_offset
        return np.array(out_offset)

    def build_state_scale(self, agent_id):
        out_scale = [1] * self.get_state_size(agent_id)
        phase_scale = 2
        out_scale[0] = phase_scale
        return np.array(out_scale)

    def get_goal_size(self, agent_id):
        return 0

    def get_action_size(self, agent_id):
        ctrl_size = 43  # numDof
        root_size = 7
        return ctrl_size - root_size

    def build_goal_norm_groups(self, agent_id):
        return np.array([])

    def build_goal_offset(self, agent_id):
        return np.array([])

    def build_goal_scale(self, agent_id):
        return np.array([])

    def build_action_offset(self, agent_id):
        out_offset = [
            0.0000000000, 0.0000000000, 0.0000000000, -0.200000000, 0.0000000000, 0.0000000000,
            0.0000000000, -0.200000000, 0.0000000000, 0.0000000000, 0.00000000, -0.2000000, 1.57000000,
            0.00000000, 0.00000000, 0.00000000, -0.2000000, 0.00000000, 0.00000000, 0.00000000,
            -0.2000000, -1.5700000, 0.00000000, 0.00000000, 0.00000000, -0.2000000, 1.57000000,
            0.00000000, 0.00000000, 0.00000000, -0.2000000, 0.00000000, 0.00000000, 0.00000000,
            -0.2000000, -1.5700000
        ]
        # see cCtCtrlUtil::BuildOffsetScalePDPrismatic and
        # see cCtCtrlUtil::BuildOffsetScalePDSpherical
        return np.array(out_offset)

    def build_action_scale(self, agent_id):
        # see cCtCtrlUtil::BuildOffsetScalePDPrismatic and
        # see cCtCtrlUtil::BuildOffsetScalePDSpherical
        out_scale = [
            0.20833333333333, 1.00000000000000, 1.00000000000000, 1.00000000000000, 0.25000000000000,
            1.00000000000000, 1.00000000000000, 1.00000000000000, 0.12077294685990, 1.00000000000000,
            1.000000000000, 1.000000000000, 0.159235668789, 0.159235668789, 1.000000000000,
            1.000000000000, 1.000000000000, 0.079617834394, 1.000000000000, 1.000000000000,
            1.000000000000, 0.159235668789, 0.120772946859, 1.000000000000, 1.000000000000,
            1.000000000000, 0.159235668789, 0.159235668789, 1.000000000000, 1.000000000000,
            1.000000000000, 0.107758620689, 1.000000000000, 1.000000000000, 1.000000000000,
            0.159235668789
        ]
        return np.array(out_scale)

    def build_action_bound_min(self, agent_id):
        # see cCtCtrlUtil::BuildBoundsPDSpherical
        out_scale = [
            -4.79999999999, -1.00000000000, -1.00000000000, -1.00000000000, -4.00000000000,
            -1.00000000000, -1.00000000000, -1.00000000000, -7.77999999999, -1.00000000000,
            -1.000000000, -1.000000000, -7.850000000, -6.280000000, -1.000000000, -1.000000000,
            -1.000000000, -12.56000000, -1.000000000, -1.000000000, -1.000000000, -4.710000000,
            -7.779999999, -1.000000000, -1.000000000, -1.000000000, -7.850000000, -6.280000000,
            -1.000000000, -1.000000000, -1.000000000, -8.460000000, -1.000000000, -1.000000000,
            -1.000000000, -4.710000000
        ]

        return out_scale

    def build_action_bound_max(self, agent_id):
        out_scale = [
            4.799999999, 1.000000000, 1.000000000, 1.000000000, 4.000000000, 1.000000000, 1.000000000,
            1.000000000, 8.779999999, 1.000000000, 1.0000000, 1.0000000, 4.7100000, 6.2800000,
            1.0000000, 1.0000000, 1.0000000, 12.560000, 1.0000000, 1.0000000, 1.0000000, 7.8500000,
            8.7799999, 1.0000000, 1.0000000, 1.0000000, 4.7100000, 6.2800000, 1.0000000, 1.0000000,
            1.0000000, 10.100000, 1.0000000, 1.0000000, 1.0000000, 7.8500000
        ]
        return out_scale

    def set_mode(self, mode):
        self._mode = mode

    def need_new_action(self, agent_id):
        if self.t >= self.needs_update_time:
            self.needs_update_time = self.t + 1. / 30.
            return True
        return False

    def record_state(self, agent_id):
        state = self._humanoid.getState()
        return np.array(state)

    def record_goal(self, agent_id):
        return np.array([])

    def calc_reward(self, agent_id):
        rewards = {}
        for i in range(self._n_clips):
            kinPose = self._humanoid.computePose(self._humanoid._frameFraction, i)
            rewards[i] = self._humanoid.getReward(kinPose, i)
        reward = self.select_reward(rewards)
        return reward

    def set_action(self, agent_id, action):
        # print("action=",)
        # for a in action:
        #  print(a)
        self.desiredPose = {}
        for i in range(self._n_clips):
            self.desiredPose[i] = self._humanoid.convertActionToPose(action, i)
            # we need the target root positon and orientation to be zero, to be compatible with deep mimic
            self.desiredPose[i][0] = 0
            self.desiredPose[i][1] = 0
            self.desiredPose[i][2] = 0
            self.desiredPose[i][3] = 0
            self.desiredPose[i][4] = 0
            self.desiredPose[i][5] = 0
            self.desiredPose[i][6] = 0
            target_pose = np.array(self.desiredPose[i])

        # print("set_action: desiredPose=", self.desiredPose)

    def log_val(self, agent_id, val):
        pass

    def update(self, timeStep):
        # print("pybullet_deep_mimic_env:update timeStep=",timeStep," t=",self.t)
        self._pybullet_client.setTimeStep(timeStep)
        self._humanoid._timeStep = timeStep
        self.timeStep = timeStep

        self.t += timeStep
        self._humanoid.setSimTime(self.t)

        if self.desiredPose:
            for i in range(self._n_clips):
                kinPose = self._humanoid.computePose(self._humanoid._frameFraction, i)
                self._humanoid.initializePose(self._humanoid._poseInterpolators[i],
                                              self._humanoid._kin_models[i],
                                              initBase=True)
            # print("desiredPositions=",self.desiredPose[i])
            maxForces = [
                0, 0, 0, 0, 0, 0, 0, 200, 200, 200, 200, 50, 50, 50, 50, 200, 200, 200, 200, 150, 90,
                90, 90, 90, 100, 100, 100, 100, 60, 200, 200, 200, 200, 150, 90, 90, 90, 90, 100, 100,
                100, 100, 60
            ]

            self._humanoid.computeAndApplyPDForces(self.desiredPose[0], maxForces=maxForces) # TODO check if this should be inside the for loop

            self._pybullet_client.stepSimulation()

    def set_sample_count(self, count):
        return

    def check_terminate(self, agent_id):
        return Env.Terminate(self.is_episode_end())

    def is_episode_end(self):
        isEnded = self._humanoid.terminates()
        # also check maximum time, 20 seconds (todo get from file)
        # print("self.t=",self.t)
        if (self.t > 20):
            isEnded = True
        return isEnded

    def check_valid_episode(self):
        # could check if limbs exceed velocity threshold
        return True

    def getKeyboardEvents(self):
        return self._pybullet_client.getKeyboardEvents()

    def isKeyTriggered(self, keys, key):
        o = ord(key)
        # print("ord=",o)
        if o in keys:
            return keys[ord(key)] & self._pybullet_client.KEY_WAS_TRIGGERED
        return False

    def select_reward(self, rewards, criterion=None):
        # todo create enum for criterion
        return max(rewards.values())



# Copyright 2017 The TensorFlow Agents Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Normalize tensors based on streaming estimates of mean and variance."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

try:
  import tensorflow.compat.v1 as tf
except Exception:
  import tensorflow as tf


class StreamingNormalize(object):
  """Normalize tensors based on streaming estimates of mean and variance."""

  def __init__(self, template, center=True, scale=True, clip=10, name='normalize'):
    """Normalize tensors based on streaming estimates of mean and variance.

    Centering the value, scaling it by the standard deviation, and clipping
    outlier values are optional.

    Args:
      template: Example tensor providing shape and dtype of the vaule to track.
      center: Python boolean indicating whether to subtract mean from values.
      scale: Python boolean indicating whether to scale values by stddev.
      clip: If and when to clip normalized values.
      name: Parent scope of operations provided by this class.
    """
    self._center = center
    self._scale = scale
    self._clip = clip
    self._name = name
    with tf.name_scope(name):
      self._count = tf.Variable(0, False)
      self._mean = tf.Variable(tf.zeros_like(template), False)
      self._var_sum = tf.Variable(tf.zeros_like(template), False)

  def transform(self, value):
    """Normalize a single or batch tensor.

    Applies the activated transformations in the constructor using current
    estimates of mean and variance.

    Args:
      value: Batch or single value tensor.

    Returns:
      Normalized batch or single value tensor.
    """
    with tf.name_scope(self._name + '/transform'):
      no_batch_dim = value.shape.ndims == self._mean.shape.ndims
      if no_batch_dim:
        # Add a batch dimension if necessary.
        value = value[None, ...]
      if self._center:
        value -= self._mean[None, ...]
      if self._scale:
        # We cannot scale before seeing at least two samples.
        value /= tf.cond(
            self._count > 1, lambda: self._std() + 1e-8, lambda: tf.ones_like(self._var_sum))[None]
      if self._clip:
        value = tf.clip_by_value(value, -self._clip, self._clip)
      # Remove batch dimension if necessary.
      if no_batch_dim:
        value = value[0]
      return tf.check_numerics(value, 'value')

  def update(self, value):
    """Update the mean and variance estimates.

    Args:
      value: Batch or single value tensor.

    Returns:
      Summary tensor.
    """
    with tf.name_scope(self._name + '/update'):
      if value.shape.ndims == self._mean.shape.ndims:
        # Add a batch dimension if necessary.
        value = value[None, ...]
      count = tf.shape(value)[0]
      with tf.control_dependencies([self._count.assign_add(count)]):
        step = tf.cast(self._count, tf.float32)
        mean_delta = tf.reduce_sum(value - self._mean[None, ...], 0)
        new_mean = self._mean + mean_delta / step
        new_mean = tf.cond(self._count > 1, lambda: new_mean, lambda: value[0])
        var_delta = (value - self._mean[None, ...]) * (value - new_mean[None, ...])
        new_var_sum = self._var_sum + tf.reduce_sum(var_delta, 0)
      with tf.control_dependencies([new_mean, new_var_sum]):
        update = self._mean.assign(new_mean), self._var_sum.assign(new_var_sum)
      with tf.control_dependencies(update):
        if value.shape.ndims == 1:
          value = tf.reduce_mean(value)
        return self._summary('value', tf.reduce_mean(value))

  def reset(self):
    """Reset the estimates of mean and variance.

    Resets the full state of this class.

    Returns:
      Operation.
    """
    with tf.name_scope(self._name + '/reset'):
      return tf.group(self._count.assign(0), self._mean.assign(tf.zeros_like(self._mean)),
                      self._var_sum.assign(tf.zeros_like(self._var_sum)))

  def summary(self):
    """Summary string of mean and standard deviation.

    Returns:
      Summary tensor.
    """
    with tf.name_scope(self._name + '/summary'):
      mean_summary = tf.cond(self._count > 0, lambda: self._summary('mean', self._mean), str)
      std_summary = tf.cond(self._count > 1, lambda: self._summary('stddev', self._std()), str)
      return tf.summary.merge([mean_summary, std_summary])

  def _std(self):
    """Computes the current estimate of the standard deviation.

    Note that the standard deviation is not defined until at least two samples
    were seen.

    Returns:
      Tensor of current variance.
    """
    variance = tf.cond(self._count > 1, lambda: self._var_sum / tf.cast(
        self._count - 1, tf.float32), lambda: tf.ones_like(self._var_sum) * float('nan'))
    # The epsilon corrects for small negative variance values caused by
    # the algorithm. It was empirically chosen to work with all environments
    # tested.
    return tf.sqrt(variance + 1e-4)

  def _summary(self, name, tensor):
    """Create a scalar or histogram summary matching the rank of the tensor.

    Args:
      name: Name for the summary.
      tensor: Tensor to summarize.

    Returns:
      Summary tensor.
    """
    if tensor.shape.ndims == 0:
      return tf.summary.scalar(name, tensor)
    else:
      return tf.summary.histogram(name, tensor)



"""Gait pattern planning module."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging
import math
from typing import Any, Sequence

import gin
import numpy as np

from pybullet_envs.minitaur.agents.baseline_controller import gait_generator

_DEFAULT_INITIAL_LEG_STATE = (
    gait_generator.LegState.STANCE,
    gait_generator.LegState.STANCE,
    gait_generator.LegState.STANCE,
    gait_generator.LegState.STANCE,
)

_NOMINAL_STANCE_DURATION = (0.25, 0.25, 0.25, 0.25)
_NOMINAL_DUTY_FACTOR = (0.6, 0.6, 0.6, 0.6)
_TROTTING_LEG_PHASE = (0, 0.5, 0.5, 0)
_NOMINAL_CONTACT_DETECTION_PHASE = 0.4


@gin.configurable
class OpenloopGaitGenerator(gait_generator.GaitGenerator):
  """Generates openloop gaits for quadruped robots.

  A flexible open-loop gait generator. Each leg has its own cycle and duty
  factor. And the state of each leg alternates between stance and swing. One can
  easily formuate a set of common quadruped gaits like trotting, pacing,
  pronking, bounding, etc by tweaking the input parameters.
  """

  def __init__(
      self,
      robot: Any,
      stance_duration: Sequence[float] = _NOMINAL_STANCE_DURATION,
      duty_factor: Sequence[float] = _NOMINAL_DUTY_FACTOR,
      initial_leg_phase: Sequence[float] = _TROTTING_LEG_PHASE,
      contact_detection_force_threshold: float = 0,
      contact_detection_phase_threshold:
      float = _NOMINAL_CONTACT_DETECTION_PHASE,
  ):
    """Initializes the class.

    Args:
      robot: A quadruped robot that at least implements the GetFootContacts API
        and num_legs property.
      stance_duration: The desired stance duration.
      duty_factor: The ratio  stance_duration / total_gait_cycle.
      initial_leg_phase: The desired initial phase [0, 1] of the legs within the
        full swing + stance cycle.
      contact_detection_force_threshold: The minimal contact force required to
        detect if a foot is in contact with the ground. For real robots this
        needs to be larger (i.e. 25 for Laikago).
      contact_detection_phase_threshold: Updates the state of each leg based on
        contact info, when the current normalized phase is greater than this
        threshold. This is essential to remove false positives in contact
        detection when phase switches. For example, a swing foot at at the
        beginning of the gait cycle might be still on the ground.
    """
    self._robot = robot
    self._stance_duration = stance_duration
    self._duty_factor = duty_factor
    self._swing_duration = np.array(stance_duration) / np.array(
        duty_factor) - np.array(stance_duration)
    if len(initial_leg_phase) != len(
        list(self._robot.urdf_loader.get_end_effector_id_dict().values())):
      raise ValueError(
          "The number of leg phases should be the same as number of legs.")
    self._initial_leg_phase = initial_leg_phase

    self._initial_leg_state = _DEFAULT_INITIAL_LEG_STATE
    self._next_leg_state = []
    # The ratio in cycle is duty factor if initial state of the leg is STANCE,
    # and 1 - duty_factory if the initial state of the leg is SWING.
    self._initial_state_ratio_in_cycle = []
    for state, duty in zip(self._initial_leg_state, duty_factor):
      assert state == gait_generator.LegState.STANCE
      self._initial_state_ratio_in_cycle.append(duty)
      self._next_leg_state.append(gait_generator.LegState.SWING)

    self._contact_detection_force_threshold = contact_detection_force_threshold
    self._contact_detection_phase_threshold = contact_detection_phase_threshold

    # The normalized phase within swing or stance duration.
    self._normalized_phase = None

    # The current leg state, when contact is considered.
    self._leg_state = None

    # The desired leg state (i.e. SWING or STANCE).
    self._desired_leg_state = None

    self.reset(0)

  def reset(self, current_time):
    # The normalized phase within swing or stance duration.
    self._normalized_phase = np.zeros(
        len(list(self._robot.urdf_loader.get_end_effector_id_dict().values())))
    self._leg_state = list(self._initial_leg_state)
    self._desired_leg_state = list(self._initial_leg_state)

  @property
  def desired_leg_state(self) -> Sequence[gait_generator.LegState]:
    """The desired leg SWING/STANCE states.

    Returns:
      The SWING/STANCE states for all legs.

    """
    return self._desired_leg_state

  @property
  def leg_state(self) -> Sequence[gait_generator.LegState]:
    """The leg state after considering contact with ground.

    Returns:
      The actual state of each leg after accounting for contacts.
    """
    return self._leg_state

  @property
  def swing_duration(self) -> Sequence[float]:
    return self._swing_duration

  @property
  def stance_duration(self) -> Sequence[float]:
    return self._stance_duration

  @property
  def normalized_phase(self) -> Sequence[float]:
    """The phase within the current swing or stance cycle.

    Reflects the leg's phase within the curren swing or stance stage. For
    example, at the end of the current swing duration, the phase will
    be set to 1 for all swing legs. Same for stance legs.

    Returns:
      Normalized leg phase for all legs.

    """
    return self._normalized_phase

  def update(self, current_time):
    contact_state = [
        np.linalg.norm(contact_force) > self._contact_detection_force_threshold
        for contact_force in self._robot.feet_contact_forces()
    ]

    for leg_id in range(
        len(list(self._robot.urdf_loader.get_end_effector_id_dict().values()))):
      # Here is the explanation behind this logic: We use the phase within the
      # full swing/stance cycle to determine if a swing/stance switch occurs
      # for a leg. The threshold value is the "initial_state_ratio_in_cycle" as
      # explained before. If the current phase is less than the initial state
      # ratio, the leg is either in the initial state or has switched back after
      # one or more full cycles.
      full_cycle_period = (
          self._stance_duration[leg_id] / self._duty_factor[leg_id])
      # To account for the non-zero initial phase, we offset the time duration
      # with the effect time contribution from the initial leg phase.
      augmented_time = current_time + self._initial_leg_phase[
          leg_id] * full_cycle_period
      phase_in_full_cycle = math.fmod(augmented_time,
                                      full_cycle_period) / full_cycle_period
      ratio = self._initial_state_ratio_in_cycle[leg_id]
      if phase_in_full_cycle < ratio:
        self._desired_leg_state[leg_id] = self._initial_leg_state[leg_id]
        self._normalized_phase[leg_id] = phase_in_full_cycle / ratio
      else:
        # A phase switch happens for this leg.
        self._desired_leg_state[leg_id] = self._next_leg_state[leg_id]
        self._normalized_phase[leg_id] = (phase_in_full_cycle - ratio) / (1 -
                                                                          ratio)

      self._leg_state[leg_id] = self._desired_leg_state[leg_id]

      # No contact detection at the beginning of each SWING/STANCE phase.
      if (self._normalized_phase[leg_id] <
          self._contact_detection_phase_threshold):
        continue
      if (self._leg_state[leg_id] == gait_generator.LegState.SWING and
          contact_state[leg_id]):
        logging.info("early touch down detected")
        self._leg_state[leg_id] = gait_generator.LegState.EARLY_CONTACT
      if (self._leg_state[leg_id] == gait_generator.LegState.STANCE and
          not contact_state[leg_id]):
        self._leg_state[leg_id] = gait_generator.LegState.LOSE_CONTACT



#The example to run the raibert controller in a Minitaur gym env.

#blaze run :minitaur_raibert_controller_example  -- --log_path=/tmp/logs

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tf.compat.v1 as tf
from pybullet_envs.minitaur.envs import minitaur_raibert_controller
from pybullet_envs.minitaur.envs import minitaur_gym_env

flags = tf.app.flags
FLAGS = tf.app.flags.FLAGS

flags.DEFINE_float("motor_kp", 1.0, "The position gain of the motor.")
flags.DEFINE_float("motor_kd", 0.015, "The speed gain of the motor.")
flags.DEFINE_float("control_latency", 0.006, "The latency between sensor measurement and action"
                   " execution the robot.")
flags.DEFINE_string("log_path", ".", "The directory to write the log file.")


def speed(t):
  max_speed = 0.35
  t1 = 3
  if t < t1:
    return t / t1 * max_speed
  else:
    return -max_speed


def main(argv):
  del argv
  env = minitaur_gym_env.MinitaurGymEnv(urdf_version=minitaur_gym_env.RAINBOW_DASH_V0_URDF_VERSION,
                                        control_time_step=0.006,
                                        action_repeat=6,
                                        pd_latency=0.0,
                                        control_latency=FLAGS.control_latency,
                                        motor_kp=FLAGS.motor_kp,
                                        motor_kd=FLAGS.motor_kd,
                                        remove_default_joint_damping=True,
                                        leg_model_enabled=False,
                                        render=True,
                                        on_rack=False,
                                        accurate_motor_model_enabled=True,
                                        log_path=FLAGS.log_path)
  env.reset()

  controller = minitaur_raibert_controller.MinitaurRaibertTrottingController(env.minitaur)

  tstart = env.minitaur.GetTimeSinceReset()
  for _ in range(1000):
    t = env.minitaur.GetTimeSinceReset() - tstart
    controller.behavior_parameters = (minitaur_raibert_controller.BehaviorParameters(
        desired_forward_speed=speed(t)))
    controller.update(t)
    env.step(controller.get_action())

  #env.close()


if __name__ == "__main__":
  tf.app.run(main)



import pybullet as p
from pdControllerExplicit import PDControllerExplicitMultiDof
from pdControllerExplicit import PDControllerExplicit
from pdControllerStable import PDControllerStable

import time

useMaximalCoordinates = False
import pybullet_data

p.connect(p.GUI)
p.setAdditionalSearchPath(pybullet_data.getDataPath())
pole = p.loadURDF("cartpole.urdf", [0, 0, 0], useMaximalCoordinates=useMaximalCoordinates)
pole2 = p.loadURDF("cartpole.urdf", [0, 1, 0], useMaximalCoordinates=useMaximalCoordinates)
pole3 = p.loadURDF("cartpole.urdf", [0, 2, 0], useMaximalCoordinates=useMaximalCoordinates)
pole4 = p.loadURDF("cartpole.urdf", [0, 3, 0], useMaximalCoordinates=useMaximalCoordinates)

exPD = PDControllerExplicitMultiDof(p)
sPD = PDControllerStable(p)

for i in range(p.getNumJoints(pole2)):
  #disable default constraint-based motors
  p.setJointMotorControl2(pole, i, p.POSITION_CONTROL, targetPosition=0, force=0)
  p.setJointMotorControl2(pole2, i, p.POSITION_CONTROL, targetPosition=0, force=0)
  p.setJointMotorControl2(pole3, i, p.POSITION_CONTROL, targetPosition=0, force=0)
  p.setJointMotorControl2(pole4, i, p.POSITION_CONTROL, targetPosition=0, force=0)

  #print("joint",i,"=",p.getJointInfo(pole2,i))

timeStepId = p.addUserDebugParameter("timeStep", 0.001, 0.1, 0.01)
desiredPosCartId = p.addUserDebugParameter("desiredPosCart", -10, 10, 2)
desiredVelCartId = p.addUserDebugParameter("desiredVelCart", -10, 10, 0)
kpCartId = p.addUserDebugParameter("kpCart", 0, 500, 1300)
kdCartId = p.addUserDebugParameter("kdCart", 0, 300, 150)
maxForceCartId = p.addUserDebugParameter("maxForceCart", 0, 5000, 1000)

textColor = [1, 1, 1]
shift = 0.05
p.addUserDebugText("explicit PD", [shift, 0, .1],
                   textColor,
                   parentObjectUniqueId=pole,
                   parentLinkIndex=1)
p.addUserDebugText("explicit PD plugin", [shift, 0, -.1],
                   textColor,
                   parentObjectUniqueId=pole2,
                   parentLinkIndex=1)
p.addUserDebugText("stablePD", [shift, 0, .1],
                   textColor,
                   parentObjectUniqueId=pole4,
                   parentLinkIndex=1)
p.addUserDebugText("position constraint", [shift, 0, -.1],
                   textColor,
                   parentObjectUniqueId=pole3,
                   parentLinkIndex=1)

desiredPosPoleId = p.addUserDebugParameter("desiredPosPole", -10, 10, 0)
desiredVelPoleId = p.addUserDebugParameter("desiredVelPole", -10, 10, 0)
kpPoleId = p.addUserDebugParameter("kpPole", 0, 500, 1200)
kdPoleId = p.addUserDebugParameter("kdPole", 0, 300, 100)
maxForcePoleId = p.addUserDebugParameter("maxForcePole", 0, 5000, 1000)

pd = p.loadPlugin("pdControlPlugin")

p.setGravity(0, 0, -10)

useRealTimeSim = False

p.setRealTimeSimulation(useRealTimeSim)

timeStep = 0.001

while p.isConnected():
  #p.getCameraImage(320,200)
  timeStep = p.readUserDebugParameter(timeStepId)
  p.setTimeStep(timeStep)

  desiredPosCart = p.readUserDebugParameter(desiredPosCartId)
  desiredVelCart = p.readUserDebugParameter(desiredVelCartId)
  kpCart = p.readUserDebugParameter(kpCartId)
  kdCart = p.readUserDebugParameter(kdCartId)
  maxForceCart = p.readUserDebugParameter(maxForceCartId)

  desiredPosPole = p.readUserDebugParameter(desiredPosPoleId)
  desiredVelPole = p.readUserDebugParameter(desiredVelPoleId)
  kpPole = p.readUserDebugParameter(kpPoleId)
  kdPole = p.readUserDebugParameter(kdPoleId)
  maxForcePole = p.readUserDebugParameter(maxForcePoleId)

  basePos, baseOrn = p.getBasePositionAndOrientation(pole)

  baseDof = 7
  taus = exPD.computePD(pole, [0, 1], [
      basePos[0], basePos[1], basePos[2], baseOrn[0], baseOrn[1], baseOrn[2], baseOrn[3],
      desiredPosCart, desiredPosPole
  ], [0, 0, 0, 0, 0, 0, 0, desiredVelCart, desiredVelPole], [0, 0, 0, 0, 0, 0, 0, kpCart, kpPole],
                        [0, 0, 0, 0, 0, 0, 0, kdCart, kdPole],
                        [0, 0, 0, 0, 0, 0, 0, maxForceCart, maxForcePole], timeStep)

  for j in [0, 1]:
    p.setJointMotorControlMultiDof(pole,
                                   j,
                                   controlMode=p.TORQUE_CONTROL,
                                   force=[taus[j + baseDof]])
  #p.setJointMotorControlArray(pole, [0,1], controlMode=p.TORQUE_CONTROL, forces=taus)

  if (pd >= 0):
    link = 0
    p.setJointMotorControl2(bodyUniqueId=pole2,
                            jointIndex=link,
                            controlMode=p.PD_CONTROL,
                            targetPosition=desiredPosCart,
                            targetVelocity=desiredVelCart,
                            force=maxForceCart,
                            positionGain=kpCart,
                            velocityGain=kdCart)
    link = 1
    p.setJointMotorControl2(bodyUniqueId=pole2,
                            jointIndex=link,
                            controlMode=p.PD_CONTROL,
                            targetPosition=desiredPosPole,
                            targetVelocity=desiredVelPole,
                            force=maxForcePole,
                            positionGain=kpPole,
                            velocityGain=kdPole)

  taus = sPD.computePD(pole4, [0, 1], [desiredPosCart, desiredPosPole],
                       [desiredVelCart, desiredVelPole], [kpCart, kpPole], [kdCart, kdPole],
                       [maxForceCart, maxForcePole], timeStep)
  #p.setJointMotorControlArray(pole4, [0,1], controlMode=p.TORQUE_CONTROL, forces=taus)
  for j in [0, 1]:
    p.setJointMotorControlMultiDof(pole4, j, controlMode=p.TORQUE_CONTROL, force=[taus[j]])

  p.setJointMotorControl2(pole3,
                          0,
                          p.POSITION_CONTROL,
                          targetPosition=desiredPosCart,
                          targetVelocity=desiredVelCart,
                          positionGain=timeStep * (kpCart / 150.),
                          velocityGain=0.5,
                          force=maxForceCart)
  p.setJointMotorControl2(pole3,
                          1,
                          p.POSITION_CONTROL,
                          targetPosition=desiredPosPole,
                          targetVelocity=desiredVelPole,
                          positionGain=timeStep * (kpPole / 150.),
                          velocityGain=0.5,
                          force=maxForcePole)

  if (not useRealTimeSim):
    p.stepSimulation()
    time.sleep(timeStep)



import sys
import subprocess
from pybullet_utils.arg_parser import ArgParser
from pybullet_utils.logger import Logger


def main():
  # Command line arguments
  args = sys.argv[1:]
  arg_parser = ArgParser()
  arg_parser.load_args(args)

  num_workers = arg_parser.parse_int('num_workers', 1)
  assert (num_workers > 0)

  Logger.print2('Running with {:d} workers'.format(num_workers))
  cmd = 'mpiexec -n {:d} python3 DeepMimic_Optimizer_multiclip.py '.format(num_workers)
  cmd += ' '.join(args)
  Logger.print2('cmd: ' + cmd)
  subprocess.call(cmd, shell=True)
  return


if __name__ == '__main__':
  main()



# Copyright 2017 The TensorFlow Agents Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Mock algorithm for testing reinforcement learning code."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tf.compat.v1 as tf


class MockAlgorithm(object):
  """Produce random actions and empty summaries."""

  def __init__(self, envs):
    """Produce random actions and empty summaries.

    Args:
      envs: List of in-graph environments.
    """
    self._envs = envs

  def begin_episode(self, unused_agent_indices):
    return tf.constant('')

  def perform(self, unused_observ):
    shape = (len(self._envs),) + self._envs[0].action_space.shape
    low = self._envs[0].action_space.low
    high = self._envs[0].action_space.high
    action = tf.random_uniform(shape) * (high - low) + low
    return action, tf.constant('')

  def experience(self, *unused_transition):
    return tf.constant('')

  def end_episode(self, unused_agent_indices):
    return tf.constant('')



r"""Laikago walking example using the locomotion controller framework.

"""

import gc
import pickle

from absl import app
from absl import flags
import numpy as np
import scipy.interpolate


from pybullet_envs.minitaur.agents.baseline_controller import locomotion_controller_setup
from pybullet_envs.minitaur.envs_v2 import env_loader
from pybullet_envs.minitaur.robots import robot_config

FLAGS = flags.FLAGS
flags.DEFINE_boolean("run_on_robot", False,
                     "whether to run in sim or on real hardware")
flags.DEFINE_boolean(
    "use_ground_truth_velocity", False,
    "whether to use a ground truth velocity estimator (available in sim)")
flags.DEFINE_enum("gait", "fast_trot",
                  ["fast_trot", "slow_trot", "walk", "stand"],
                  "The gait pattern to use")
flags.DEFINE_boolean(
    "use_keyboard_control", False,
    "whether to use a keyboard to control or demo speed profile.")
flags.DEFINE_string("log_path", None, "Path to save robot logs")
flags.DEFINE_boolean("add_random_push", False,
                     "whether to add random push to the robot in simulation")

_MAX_TIME_SECONDS = 100


def _load_config(render=True, run_on_robot=False):
  """Builds the environment for the quadruped robot.

  Args:
    render: Enable/disable rendering.
    run_on_robot: Whether deploy to robot or run in sim.
  """
  if run_on_robot:
    locomotion_controller_setup.load_real_config()
  else:
    locomotion_controller_setup.load_sim_config(render)
    if FLAGS.add_random_push:
      locomotion_controller_setup.add_random_push_config()


def _generate_example_linear_angular_speed(t):
  """Creates an example speed profile based on time for demo purpose."""
  vx = 0.1
  vy = 0.1
  wz = 0.3
  time_points = (0, 4, 7, 11, 13, 15, 17, 19, 100)
  speed_points = ((0, 0, 0, 0), (vx, 0, 0, 0), (-vx, 0, 0, 0), (0, -vy, 0, 0),
                  (0, vy, 0, 0), (0, 0, 0, wz), (0, 0, 0, -wz), (0, 0, 0, 0),
                  (0, 0, 0, 0))

  speed = scipy.interpolate.interp1d(
      time_points,
      speed_points,
      kind="previous",
      fill_value="extrapolate",
      axis=0)(
          t)

  return speed[0:3], speed[3]


def _update_speed_from_kb(kb, lin_speed, ang_speed):
  """Updates the controller behavior parameters."""
  if kb.is_keyboard_hit():
    c = kb.get_input_character()
    if c == "w":
      lin_speed += np.array((0.05, 0, 0))
    if c == "s":
      lin_speed += np.array((-0.05, 0, 0))
    if c == "q":
      ang_speed += 0.1
    if c == "e":
      ang_speed += -0.1
    if c == "a":
      lin_speed += np.array((0, 0.05, 0))
    if c == "d":
      lin_speed += np.array((0, -0.05, 0))
    if c == "r":
      lin_speed = np.array([0.0, 0.0, 0.0])
      ang_speed = 0.0

    lin_speed[0] = np.clip(lin_speed[0], -0.2, 0.4)
    lin_speed[1] = np.clip(lin_speed[1], -0.2, 0.2)
    ang_speed = np.clip(ang_speed, -0.3, 0.3)
    print("desired speed: ", lin_speed, ang_speed)

  return lin_speed, ang_speed


def _update_controller_params(controller, lin_speed, ang_speed):
  controller.swing_leg_controller.desired_speed = lin_speed
  controller.swing_leg_controller.desired_twisting_speed = ang_speed
  controller.stance_leg_controller.desired_speed = lin_speed
  controller.stance_leg_controller.desired_twisting_speed = ang_speed


def _run_example(max_time=_MAX_TIME_SECONDS,
                 run_on_robot=False,
                 use_keyboard=False):
  """Runs the locomotion controller example."""
  if use_keyboard:
    kb = keyboard_utils.KeyboardInput()

  env = env_loader.load()
  env.reset()

  # To mitigate jittering from the python
  gc.collect()

  # Wait for the robot to be placed properly.
  if run_on_robot:
    input("Press Enter to continue when robot is ready.")

  lin_speed = np.array([0.0, 0.0, 0.0])
  ang_speed = 0.0

  controller = locomotion_controller_setup.setup_controller(
      env.robot, FLAGS.gait, run_on_robot, FLAGS.use_ground_truth_velocity)
  controller.reset()

  loop_start_time = env.get_time_since_reset()
  loop_elapsed_time = 0
  robot_log = {
      "timestamps": [],
      "motor_angles": [],
      "motor_velocities": [],
      "base_velocities": [],
      "foot_positions": [],
      "base_rollpitchyaw": [],
      "base_angular_velocities": [],
      "actions": []
  }
  try:
    while loop_elapsed_time < max_time:
      #if use_keyboard:
      #  lin_speed, ang_speed = _update_speed_from_kb(kb, lin_speed, ang_speed)
      #else:
      lin_speed, ang_speed = _generate_example_linear_angular_speed(
            loop_elapsed_time)

      # Needed before every call to get_action().
      _update_controller_params(controller, lin_speed, ang_speed)
      controller.update()
      hybrid_action = controller.get_action()

      # Log the robot data.
      robot_log["timestamps"].append(env.robot.GetTimeSinceReset())
      robot_log["motor_angles"].append(env.robot.motor_angles)
      robot_log["motor_velocities"].append(env.robot.motor_velocities)
      robot_log["base_velocities"].append(
          controller.state_estimator.com_velocity_body_yaw_aligned_frame)
      robot_log["foot_positions"].append(env.robot.foot_positions())
      robot_log["base_rollpitchyaw"].append(env.robot.base_roll_pitch_yaw)
      robot_log["base_angular_velocities"].append(
          env.robot.base_roll_pitch_yaw_rate)
      robot_log["actions"].append(hybrid_action)

      env.step(hybrid_action)
      loop_elapsed_time = env.get_time_since_reset() - loop_start_time

  finally:
    if FLAGS.run_on_robot:
      # Apply zero torques to the robot.
      env.robot.apply_action(
          [0] * env.robot.num_motors,
          motor_control_mode=robot_config.MotorControlMode.TORQUE)
    if FLAGS.log_path:
      pickle.dump(robot_log, gfile.Open(FLAGS.log_path + "/robot.log", "wb"))


def main(argv):
  del argv
  _load_config(render=True, run_on_robot=FLAGS.run_on_robot)
  _run_example(
      run_on_robot=FLAGS.run_on_robot, use_keyboard=FLAGS.use_keyboard_control)


if __name__ == "__main__":
  app.run(main)



"""Gym environment of minitaur which randomize the terrain at each reset."""
import math
import os

import numpy as np
#from google3.pyglib import flags
#from google3.pyglib import gfile
from pybullet_envs.minitaur.envs import minitaur
from pybullet_envs.minitaur.envs import minitaur_gym_env

#flags.DEFINE_string(
#    'terrain_dir', '/cns/od-d/home/brain-minitaur/terrain_obj',
#    'The directory which contains terrain obj files to be used.')
#flags.DEFINE_string('storage_dir', '/tmp',
#                    'The full path to the temporary directory to be used.')
#FLAGS = flags.FLAGS


class MinitaurRandomizeTerrainGymEnv(minitaur_gym_env.MinitaurGymEnv):
  """The gym environment for the minitaur with randomized terrain.

  It simulates a minitaur (a quadruped robot) on a randomized terrain. The state
  space include the angles, velocities and torques for all the motors and the
  action space is the desired motor angle for each motor. The reward function is
  based on how far the minitaur walks in 1000 steps and penalizes the energy
  expenditure.

  """

  def reset(self):
    self._pybullet_client.resetSimulation()
    self._pybullet_client.setPhysicsEngineParameter(
        numSolverIterations=self._num_bullet_solver_iterations)
    self._pybullet_client.setTimeStep(self._time_step)
    terrain_visual_shape_id = -1
    terrain_mass = 0
    terrain_position = [0, 0, 0]
    terrain_orientation = [0, 0, 0, 1]
    terrain_file_name = self.load_random_terrain(FLAGS.terrain_dir)
    terrain_collision_shape_id = self._pybullet_client.createCollisionShape(
        shapeType=self._pybullet_client.GEOM_MESH,
        fileName=terrain_file_name,
        flags=1,
        meshScale=[0.5, 0.5, 0.5])
    self._pybullet_client.createMultiBody(terrain_mass, terrain_collision_shape_id,
                                          terrain_visual_shape_id, terrain_position,
                                          terrain_orientation)
    self._pybullet_client.setGravity(0, 0, -10)
    self.minitaur = (minitaur.Minitaur(pybullet_client=self._pybullet_client,
                                       urdf_root=self._urdf_root,
                                       time_step=self._time_step,
                                       self_collision_enabled=self._self_collision_enabled,
                                       motor_velocity_limit=self._motor_velocity_limit,
                                       pd_control_enabled=self._pd_control_enabled,
                                       on_rack=self._on_rack))
    self._last_base_position = [0, 0, 0]
    for _ in xrange(100):
      if self._pd_control_enabled:
        self.minitaur.ApplyAction([math.pi / 2] * 8)
      self._pybullet_client.stepSimulation()
    return self._get_observation()

  def load_random_terrain(self, terrain_dir):
    """Load a random terrain obj file.

    Args:
      terrain_dir: The directory which contains terrain obj files to be used.

    Returns:
      terrain_file_name_complete: The complete terrain obj file name in the
      local directory.
    """
    terrain_file_names_all = gfile.ListDir(terrain_dir)
    terrain_file_name = np.random.choice(terrain_file_names_all)
    asset_source = os.path.join(terrain_dir, terrain_file_name)
    asset_destination = os.path.join(FLAGS.storage_dir, terrain_file_name)
    gfile.Copy(asset_source, asset_destination, overwrite=True)
    terrain_file_name_complete = os.path.join(FLAGS.storage_dir, terrain_file_name)
    return terrain_file_name_complete
